# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: deeplab2/trainer.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='deeplab2/trainer.proto',
    package='deeplab2',
    syntax='proto2',
    serialized_options=b'P\001',
    serialized_pb=b'\n\x16\x64\x65\x65plab2/trainer.proto\x12\x08\x64\x65\x65plab2\"\xe4\x03\n\rSolverOptions\x12 \n\x12use_sync_batchnorm\x18\x01 \x01(\x08:\x04true\x12 \n\x12\x62\x61tchnorm_momentum\x18\x0e \x01(\x02:\x04\x30.99\x12 \n\x11\x62\x61tchnorm_epsilon\x18\x0f \x01(\x02:\x05\x30.001\x12\x1d\n\x0flearning_policy\x18\x02 \x01(\t:\x04poly\x12!\n\x12\x62\x61se_learning_rate\x18\x03 \x01(\x02:\x05\x30.001\x12 \n\x13poly_learning_power\x18\x04 \x01(\x02:\x03\x30.9\x12!\n\x16poly_end_learning_rate\x18\x05 \x01(\x02:\x01\x30\x12\x17\n\x0cwarmup_steps\x18\x06 \x01(\x05:\x01\x30\x12\x17\n\toptimizer\x18\x07 \x01(\t:\x04\x61\x64\x61m\x12\x17\n\x0cweight_decay\x18\x08 \x01(\x02:\x01\x30\x12$\n\x15use_gradient_clipping\x18\t \x01(\x08:\x05\x66\x61lse\x12\x1e\n\x12\x63lip_gradient_norm\x18\n \x01(\x02:\x02\x31\x30\x12\'\n\x18training_number_of_steps\x18\x0b \x01(\x05:\x05\x36\x30\x30\x30\x30\x12,\n!backbone_learning_rate_multiplier\x18\x10 \x01(\x02:\x01\x31\"\xbb\x05\n\x0bLossOptions\x12>\n\rsemantic_loss\x18\x01 \x01(\x0b\x32\'.deeplab2.LossOptions.SingleLossOptions\x12<\n\x0b\x63\x65nter_loss\x18\x02 \x01(\x0b\x32\'.deeplab2.LossOptions.SingleLossOptions\x12@\n\x0fregression_loss\x18\x03 \x01(\x0b\x32\'.deeplab2.LossOptions.SingleLossOptions\x12<\n\x0bmotion_loss\x18\x04 \x01(\x0b\x32\'.deeplab2.LossOptions.SingleLossOptions\x12\x45\n\x14next_regression_loss\x18\x05 \x01(\x0b\x32\'.deeplab2.LossOptions.SingleLossOptions\x12>\n\rpq_style_loss\x18\x06 \x01(\x0b\x32\'.deeplab2.LossOptions.SingleLossOptions\x12K\n\x1amask_id_cross_entropy_loss\x18\x07 \x01(\x0b\x32\'.deeplab2.LossOptions.SingleLossOptions\x12M\n\x1cinstance_discrimination_loss\x18\x08 \x01(\x0b\x32\'.deeplab2.LossOptions.SingleLossOptions\x12;\n\ndepth_loss\x18\t \x01(\x0b\x32\'.deeplab2.LossOptions.SingleLossOptions\x1aN\n\x11SingleLossOptions\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\x06weight\x18\x02 \x01(\x02:\x01\x31\x12\x18\n\rtop_k_percent\x18\x03 \x01(\x02:\x01\x31\"\xfa\x01\n\x0eTrainerOptions\x12\"\n\x17num_checkpoints_to_keep\x18\x01 \x01(\x05:\x01\x35\x12$\n\x16save_checkpoints_steps\x18\x02 \x01(\x05:\x04\x31\x30\x30\x30\x12\"\n\x14save_summaries_steps\x18\x03 \x01(\x05:\x04\x31\x30\x30\x30\x12\x1c\n\x0esteps_per_loop\x18\x04 \x01(\x05:\x04\x31\x30\x30\x30\x12+\n\x0closs_options\x18\x05 \x01(\x0b\x32\x15.deeplab2.LossOptions\x12/\n\x0esolver_options\x18\x06 \x01(\x0b\x32\x17.deeplab2.SolverOptionsB\x02P\x01'
)

_SOLVEROPTIONS = _descriptor.Descriptor(
    name='SolverOptions',
    full_name='deeplab2.SolverOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='use_sync_batchnorm', full_name='deeplab2.SolverOptions.use_sync_batchnorm', index=0,
            number=1, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='batchnorm_momentum', full_name='deeplab2.SolverOptions.batchnorm_momentum', index=1,
            number=14, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.99),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='batchnorm_epsilon', full_name='deeplab2.SolverOptions.batchnorm_epsilon', index=2,
            number=15, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.001),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='learning_policy', full_name='deeplab2.SolverOptions.learning_policy', index=3,
            number=2, type=9, cpp_type=9, label=1,
            has_default_value=True, default_value=b"poly".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='base_learning_rate', full_name='deeplab2.SolverOptions.base_learning_rate', index=4,
            number=3, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.001),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='poly_learning_power', full_name='deeplab2.SolverOptions.poly_learning_power', index=5,
            number=4, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.9),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='poly_end_learning_rate', full_name='deeplab2.SolverOptions.poly_end_learning_rate', index=6,
            number=5, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='warmup_steps', full_name='deeplab2.SolverOptions.warmup_steps', index=7,
            number=6, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='optimizer', full_name='deeplab2.SolverOptions.optimizer', index=8,
            number=7, type=9, cpp_type=9, label=1,
            has_default_value=True, default_value=b"adam".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='weight_decay', full_name='deeplab2.SolverOptions.weight_decay', index=9,
            number=8, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='use_gradient_clipping', full_name='deeplab2.SolverOptions.use_gradient_clipping', index=10,
            number=9, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='clip_gradient_norm', full_name='deeplab2.SolverOptions.clip_gradient_norm', index=11,
            number=10, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(10),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='training_number_of_steps', full_name='deeplab2.SolverOptions.training_number_of_steps', index=12,
            number=11, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=60000,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='backbone_learning_rate_multiplier',
            full_name='deeplab2.SolverOptions.backbone_learning_rate_multiplier', index=13,
            number=16, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=37,
    serialized_end=521,
)

_LOSSOPTIONS_SINGLELOSSOPTIONS = _descriptor.Descriptor(
    name='SingleLossOptions',
    full_name='deeplab2.LossOptions.SingleLossOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='name', full_name='deeplab2.LossOptions.SingleLossOptions.name', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='weight', full_name='deeplab2.LossOptions.SingleLossOptions.weight', index=1,
            number=2, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='top_k_percent', full_name='deeplab2.LossOptions.SingleLossOptions.top_k_percent', index=2,
            number=3, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=1145,
    serialized_end=1223,
)

_LOSSOPTIONS = _descriptor.Descriptor(
    name='LossOptions',
    full_name='deeplab2.LossOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='semantic_loss', full_name='deeplab2.LossOptions.semantic_loss', index=0,
            number=1, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='center_loss', full_name='deeplab2.LossOptions.center_loss', index=1,
            number=2, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='regression_loss', full_name='deeplab2.LossOptions.regression_loss', index=2,
            number=3, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='motion_loss', full_name='deeplab2.LossOptions.motion_loss', index=3,
            number=4, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='next_regression_loss', full_name='deeplab2.LossOptions.next_regression_loss', index=4,
            number=5, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='pq_style_loss', full_name='deeplab2.LossOptions.pq_style_loss', index=5,
            number=6, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='mask_id_cross_entropy_loss', full_name='deeplab2.LossOptions.mask_id_cross_entropy_loss', index=6,
            number=7, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='instance_discrimination_loss', full_name='deeplab2.LossOptions.instance_discrimination_loss', index=7,
            number=8, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='depth_loss', full_name='deeplab2.LossOptions.depth_loss', index=8,
            number=9, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[_LOSSOPTIONS_SINGLELOSSOPTIONS, ],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=524,
    serialized_end=1223,
)

_TRAINEROPTIONS = _descriptor.Descriptor(
    name='TrainerOptions',
    full_name='deeplab2.TrainerOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='num_checkpoints_to_keep', full_name='deeplab2.TrainerOptions.num_checkpoints_to_keep', index=0,
            number=1, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=5,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='save_checkpoints_steps', full_name='deeplab2.TrainerOptions.save_checkpoints_steps', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=1000,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='save_summaries_steps', full_name='deeplab2.TrainerOptions.save_summaries_steps', index=2,
            number=3, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=1000,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='steps_per_loop', full_name='deeplab2.TrainerOptions.steps_per_loop', index=3,
            number=4, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=1000,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='loss_options', full_name='deeplab2.TrainerOptions.loss_options', index=4,
            number=5, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='solver_options', full_name='deeplab2.TrainerOptions.solver_options', index=5,
            number=6, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=1226,
    serialized_end=1476,
)

_LOSSOPTIONS_SINGLELOSSOPTIONS.containing_type = _LOSSOPTIONS
_LOSSOPTIONS.fields_by_name['semantic_loss'].message_type = _LOSSOPTIONS_SINGLELOSSOPTIONS
_LOSSOPTIONS.fields_by_name['center_loss'].message_type = _LOSSOPTIONS_SINGLELOSSOPTIONS
_LOSSOPTIONS.fields_by_name['regression_loss'].message_type = _LOSSOPTIONS_SINGLELOSSOPTIONS
_LOSSOPTIONS.fields_by_name['motion_loss'].message_type = _LOSSOPTIONS_SINGLELOSSOPTIONS
_LOSSOPTIONS.fields_by_name['next_regression_loss'].message_type = _LOSSOPTIONS_SINGLELOSSOPTIONS
_LOSSOPTIONS.fields_by_name['pq_style_loss'].message_type = _LOSSOPTIONS_SINGLELOSSOPTIONS
_LOSSOPTIONS.fields_by_name['mask_id_cross_entropy_loss'].message_type = _LOSSOPTIONS_SINGLELOSSOPTIONS
_LOSSOPTIONS.fields_by_name['instance_discrimination_loss'].message_type = _LOSSOPTIONS_SINGLELOSSOPTIONS
_LOSSOPTIONS.fields_by_name['depth_loss'].message_type = _LOSSOPTIONS_SINGLELOSSOPTIONS
_TRAINEROPTIONS.fields_by_name['loss_options'].message_type = _LOSSOPTIONS
_TRAINEROPTIONS.fields_by_name['solver_options'].message_type = _SOLVEROPTIONS
DESCRIPTOR.message_types_by_name['SolverOptions'] = _SOLVEROPTIONS
DESCRIPTOR.message_types_by_name['LossOptions'] = _LOSSOPTIONS
DESCRIPTOR.message_types_by_name['TrainerOptions'] = _TRAINEROPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SolverOptions = _reflection.GeneratedProtocolMessageType('SolverOptions', (_message.Message,), {
    'DESCRIPTOR': _SOLVEROPTIONS,
    '__module__': 'deeplab2.trainer_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.SolverOptions)
})
_sym_db.RegisterMessage(SolverOptions)

LossOptions = _reflection.GeneratedProtocolMessageType('LossOptions', (_message.Message,), {

    'SingleLossOptions': _reflection.GeneratedProtocolMessageType('SingleLossOptions', (_message.Message,), {
        'DESCRIPTOR': _LOSSOPTIONS_SINGLELOSSOPTIONS,
        '__module__': 'deeplab2.trainer_pb2'
        # @@protoc_insertion_point(class_scope:deeplab2.LossOptions.SingleLossOptions)
    })
    ,
    'DESCRIPTOR': _LOSSOPTIONS,
    '__module__': 'deeplab2.trainer_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.LossOptions)
})
_sym_db.RegisterMessage(LossOptions)
_sym_db.RegisterMessage(LossOptions.SingleLossOptions)

TrainerOptions = _reflection.GeneratedProtocolMessageType('TrainerOptions', (_message.Message,), {
    'DESCRIPTOR': _TRAINEROPTIONS,
    '__module__': 'deeplab2.trainer_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.TrainerOptions)
})
_sym_db.RegisterMessage(TrainerOptions)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3
r"""
CommandLine:
    python ~/code/watch/dev/validate_annotation_schemas.py

Prereq:


    TEST_DPATH=$HOME/tmp/test_smart_schema
    mkdir -p $TEST_DPATH
    cd $TEST_DPATH
    git clone git@smartgitlab.com:TE/annotations.git
    git clone git@smartgitlab.com:infrastructure/docs.git

    pip install jsonschema ubelt -U


SeeAlso:
    ~/code/watch/geowatch/geoannots/geomodels.py
    ~/code/watch/geowatch/cli/validate_annotation_schemas.py
    ~/code/watch/geowatch/cli/fix_region_models.py


References:
    https://smartgitlab.com/TE/annotations/-/issues/17
    https://smartgitlab.com/TE/standards/-/snippets/18


Example:

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)

    python -m geowatch.cli.validate_annotation_schemas \
        --site_models="$DVC_DATA_DPATH"/annotations/drop7/site_models \
        --region_models="$DVC_DATA_DPATH"/annotations/drop7/region_models

    python -m geowatch.cli.validate_annotation_schemas \
        --site_models="<path-to-site-models>" \
        --region_models="<path-to-region-models>"

    python -m geowatch.cli.validate_annotation_schemas \
        --region_models="$DVC_DATA_DPATH"/annotations/drop6/region_models/AE_C001.geojson
"""

import ubelt as ub
import jsonschema
import scriptconfig as scfg


class ValidateAnnotationConfig(scfg.DataConfig):
    """
    Validate the site / region model schemas
    """
    __command__ = 'site_validate'
    __alias__ = ['validate_sites']

    models = scfg.Value(None, help='site OR region models coercables (the script will attempt to distinguish them)', nargs='+', position=1)

    site_models = scfg.Value(None, nargs='+', help='coercable site models')

    region_models = scfg.Value(None, nargs='+', help='coercable region models')

    io_workers = scfg.Value('avail', help='number of workers for parallel io')

    fixup = scfg.Value(False, isflag=True, help='if True then run fixups before validation to check if the fixed version works')

    strict = scfg.Value(False, help='if True use the strict schema (i.e. enforces IARPA naming)')


def main(cmdline=1, **kwargs):
    """
    Example:
        >>> import ubelt as ub
        >>> import sys, ubelt
        >>> from geowatch.cli.validate_annotation_schemas import ValidateAnnotationConfig
        >>> from geowatch.geoannots import geomodels
        >>> dpath = ub.Path.appdir('geowatch', 'tests', 'test_validate_geoannot_schema')
        >>> dpath.ensuredir()
        >>> region, sites = geomodels.RegionModel.random(with_sites=True)
        >>> region_fpath = dpath / (region.region_id + '.geojson')
        >>> region_fpath.write_text(region.dumps())
        >>> for site in sites:
        >>>     site_fpath = dpath / (site.site_id + '.geojson')
        >>>     site_fpath.write_text(site.dumps())
        >>> kwargs = {
        >>>     'models': str(dpath / '*.geojson')
        >>> }
        >>> cmdline = 0
        >>> ValidateAnnotationConfig.main(cmdline=cmdline, **kwargs)
    """
    config = ValidateAnnotationConfig.cli(data=kwargs, cmdline=cmdline)
    from kwgis.utils import util_gis
    import rich
    rich.print(ub.urepr(config))

    from geowatch.geoannots import geomodels
    from kwutil import util_parallel

    io_workers = util_parallel.coerce_num_workers(config['io_workers'])

    if config.models:
        if config.site_models:
            raise ValueError('the models and site_models arguments are mutex')
        if config.region_models:
            raise ValueError('the models and region_models arguments are mutex')
        model_infos = list(util_gis.coerce_geojson_datas(config.models, format='json', workers=io_workers))
        site_model_infos = []
        region_model_infos = []
        for model_info in model_infos:
            model_data = model_info.pop('data')
            model = geomodels.coerce_site_or_region_model(model_data)
            model_info['model'] = model
            if isinstance(model, geomodels.SiteModel):
                site_model_infos.append(model_info)
            elif isinstance(model, geomodels.RegionModel):
                region_model_infos.append(model_info)
            else:
                raise AssertionError
    else:
        site_model_infos = list(util_gis.coerce_geojson_datas(config['site_models'], format='json', workers=io_workers))
        region_model_infos = list(util_gis.coerce_geojson_datas(config['region_models'], format='json', workers=io_workers))

        for model_info in region_model_infos + site_model_infos:
            model_data = model_info.pop('data')
            model = geomodels.coerce_site_or_region_model(model_data)
            model_info['model'] = model

    if config.fixup:
        for info in region_model_infos:
            info['model'].fixup()

        for info in site_model_infos:
            info['model'].fixup()

    print('Validate Data Content')
    validate_data_contents(region_model_infos, site_model_infos)

    print('Validate Schemas')
    validate_schemas(region_model_infos, site_model_infos,
                     strict=config.strict)


def validate_data_contents(region_model_infos, site_model_infos):
    """
    Content validation (i.e. dates look sane)
    """
    # Finish reading region models and build reports
    region_reports = []
    for region_info in region_model_infos:
        region_model = region_info['model']
        region_fpath = region_info['fpath']
        region_df = region_model.pandas()
        region_report = validate_region_model_content(region_df, region_fpath)
        region_reports.append(region_report)
    region_reports = sorted(region_reports, key=lambda x: (x['num_errors'], x['region_id']))
    print('region_reports = {}'.format(ub.urepr(region_reports, sort=0, nl=-1)).replace('\'', '"').replace('None', 'null'))

    # Finish reading site models, build reports while this is happening.
    site_reports = []
    for site_info in site_model_infos:
        site_model = site_info['model']
        site_fpath = site_info['fpath']
        site_df = site_model.pandas()
        site_report = validate_site_content(site_df, site_fpath)
        site_reports.append(site_report)

    region_id_to_site_reports = ub.group_items(site_reports, lambda x: x['region_id'])
    region_id_to_report_group = {}
    for region_id, sub_reports in region_id_to_site_reports.items():
        total_errors = sum([r['num_errors'] for r in sub_reports])
        invalid_site_reports = [r for r in sub_reports if r['num_errors']]
        for r in invalid_site_reports:
            r.pop('region_id', None)
        region_id_to_report_group[region_id] = {
            'region_id': region_id,
            'num_errors': total_errors,
            'valid_sites': len(sub_reports) - len(invalid_site_reports),
            'invalid_site_reports': invalid_site_reports,
        }

    grouped_site_reports = sorted(region_id_to_report_group.values(), key=lambda x: (x['num_errors'], x['region_id']))
    print('site_reports = {}'.format(ub.urepr(grouped_site_reports, sort=0, nl=-1)).replace('\'', '"').replace('None', 'null'))


def validate_schemas(region_model_infos, site_model_infos, strict=False):

    region_errors = []
    prog = ub.ProgIter(region_model_infos, desc='check region models')
    for region_model_info in prog:
        region_model_fpath = region_model_info['fpath']
        region_model = region_model_info['model']
        try:
            region_model.validate(verbose=0, strict=strict)
        except jsonschema.ValidationError as ex:
            error_info = {
                'type': 'region_model_error',
                'name': region_model_fpath.stem,
                'ex': ex,
            }
            region_errors.append(error_info)
            prog.set_description(f'check region models, errors: {len(region_errors)}')
        except Exception as ex:
            error_info = {
                'type': 'region_model_error',
                'name': region_model_fpath.stem,
                'ex': ex,
            }
            region_errors.append(error_info)
            prog.set_description(f'check region models, errors: {len(region_errors)}')

    site_errors = []
    prog = ub.ProgIter(site_model_infos, desc='check site models')
    for site_model_info in prog:
        site_model_fpath = site_model_info['fpath']
        site_model = site_model_info['model']
        try:
            site_model.validate(verbose=0, strict=strict)
            # site_model._validate_parts(strict=strict)
        except jsonschema.ValidationError as ex:
            error_info = {
                'type': 'site_model_error',
                'name': site_model_fpath.stem,
                'ex': ex,
            }
            site_errors.append(error_info)
            prog.set_description(f'check site models, errors: {len(site_errors)}')
        except Exception as ex:
            error_info = {
                'type': 'site_model_error',
                'name': site_model_fpath.stem,
                'ex': ex,
            }
            site_errors.append(error_info)
            prog.set_description(f'check site models, errors: {len(site_errors)}')

    print('site_errors = {}'.format(ub.urepr(site_errors, nl=1)))
    print('region_errors = {}'.format(ub.urepr(region_errors, nl=1)))

    print(f'{len(site_errors)} / {len(site_model_infos)} site model errors')
    print(f'{len(region_errors)} / {len(region_model_infos)} region model errors')


def validate_region_model_content(region_df, fpath):
    # import pandas as pd
    import os
    from kwutil import util_time
    is_region = region_df['type'] == 'region'
    region_part = region_df[is_region]
    assert len(region_part) == 1, 'must have exactly one region in each region file'
    assert region_part['region_id'].apply(lambda x: x is not None).all(), 'regions must have region ids'

    region_rel_fpath = fpath.relative_to(fpath.parent.parent.parent)

    region_row = region_part.iloc[0]
    region_id = region_row['region_id']
    region_stem = fpath.stem

    errors = []
    region_report = {
        'region_id': region_id,
        'rel_fpath': os.fspath(region_rel_fpath),
        'errors': errors,
    }

    if region_id != region_stem:
        errors.append({
            'description': 'A region file name does not match its region id',
            'offending_info': {
                'region_id': region_id,
                'filename': fpath.name,
            }
        })

    dummy_start_date = '1970-01-01'  # hack, could be more robust here
    dummy_end_date = '2101-01-01'

    class AbortCheck(Exception):
        ...

    # Hack to set all region-ids
    sites_part = region_df[~is_region]
    try:
        if not (sites_part['type'] == 'site_summary').all():
            bad_types = sites_part['type'].unique()
            errors.append({
                'description': f'rest of data must be site summaries, but got: {bad_types}'
            })
            raise AbortCheck

        region_start_date = region_row['start_date'] or dummy_start_date
        region_end_date = region_row['end_date'] or dummy_end_date

        region_start_datetime = util_time.coerce_datetime(region_start_date)
        region_end_datetime = util_time.coerce_datetime(region_end_date)
        if region_end_datetime < region_start_datetime:
            errors.append(f'Bad region dates: {region_start_datetime=}, {region_end_datetime=}')

        # Check datetime errors
        sitesum_start_dates = sites_part['start_date'].apply(lambda x: util_time.coerce_datetime(x or region_start_date))
        sitesum_end_dates = sites_part['end_date'].apply(lambda x: util_time.coerce_datetime(x or region_end_date))
        has_bad_time_range = sitesum_start_dates > sitesum_end_dates

        bad_date_rows = sites_part[has_bad_time_range]
        if len(bad_date_rows):
            bad_row_info = bad_date_rows[['site_id', 'start_date', 'end_date', 'originator']].to_dict('records')
            errors.append({
                'description': f'Site summary rows with start_dates > end_date or outside of region start/end dates ({region_start_date} - {region_end_date})',
                'offending_rows': bad_row_info,
            })
    except AbortCheck:
        ...
    except Exception as ex:
        errors.append({
            'description': 'Unknown error {}'.format(repr(ex))
        })

    if not region_report['errors']:
        region_report['errors'] = None
    region_report['num_errors'] = len(errors)
    return region_report


def validate_site_content(site_df, site_fpath):
    from kwutil import util_time
    import os
    import numpy as np
    dummy_start_date = '1970-01-01'  # hack, could be more robust here
    dummy_end_date = '2101-01-01'
    first = site_df.iloc[0]
    rest = site_df.iloc[1:]
    assert first['type'] == 'site', 'first row must have type of site'
    assert first['region_id'] is not None, 'first row must have a region id'
    assert rest['type'].apply(lambda x: x == 'observation').all(), (
        'rest of row must have type observation')
    assert rest['region_id'].isna().all(), (
        'rest of row must have region_id=None')

    site_start_date = first['start_date'] or dummy_start_date
    site_end_date = first['end_date'] or dummy_end_date
    site_start_datetime = util_time.coerce_datetime(site_start_date)
    site_end_datetime = util_time.coerce_datetime(site_end_date)

    rel_fpath = site_fpath.relative_to(site_fpath.parent.parent.parent)

    errors = []
    site_report = {
        'region_id': first['region_id'],
        'site_id': first['site_id'],
        'rel_fpath': os.fspath(rel_fpath),
        'errors': errors,
    }

    if site_end_datetime is not None and site_start_datetime is not None:
        if site_end_datetime < site_start_datetime:
            offending = first[['site_id', 'start_date', 'end_date', 'originator']].to_dict()
            errors.append({
                'description': 'Site summary row has a start_date > end_date',
                'offending_info': offending,
            })

    # Check datetime errors in observations
    null_obs_sites = []
    try:
        obs_dates = [None if x is None else util_time.coerce_datetime(x) for x in rest['observation_date']]
        obs_isvalid = [x is not None for x in obs_dates]
        valid_obs_dates = list(ub.compress(obs_dates, obs_isvalid))
        if not all(valid_obs_dates):
            null_obs_sites.append(first[['site_id', 'status']].to_dict())
        valid_deltas = np.array([d.total_seconds() for d in np.diff(valid_obs_dates)])
        if not (valid_deltas >= 0).all():
            errors.append({
                'description': 'observations are not sorted temporally',
                'offending_info': valid_obs_dates,
            })
    except AssertionError as ex:
        print('ex = {!r}'.format(ex))
        print(site_df)
        raise

    if null_obs_sites:
        errors.append({
            'description': 'sites contained null observations',
            'offending_info': null_obs_sites,
        })

    if not site_report['errors']:
        site_report['errors'] = None
    site_report['num_errors'] = len(errors)

    return site_report


__cli__ = ValidateAnnotationConfig
__cli__.main = main


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import os
import errno
import json
from lydorn_utils import print_utils


def module_exists(module_name):
    try:
        __import__(module_name)
    except ImportError:
        return False
    else:
        return True


def choose_first_existing_path(path_list, return_tried_paths=False):
    paths_tried = []
    for path in path_list:
        path = os.path.expanduser(path)
        paths_tried.append(path)
        if os.path.exists(path):
            if return_tried_paths:
                return path, paths_tried
            else:
                return path

    if return_tried_paths:
        return None, paths_tried
    else:
        return None


def get_display_availability():
    return "DISPLAY" in os.environ


def get_filepaths(dir_path, endswith_str="", startswith_str="", not_endswith_str=None, not_startswith_str=None):
    if os.path.isdir(dir_path):
        image_filepaths = []
        for path, dnames, fnames in os.walk(dir_path):
            fnames = sorted(fnames)
            image_filepaths.extend([os.path.join(path, x) for x in fnames
                                    if x.endswith(endswith_str)
                                    and x.startswith(startswith_str)
                                    and (not_endswith_str is None or not x.endswith(not_endswith_str))
                                    and (not_startswith_str is None or not x.startswith(not_startswith_str))])
        return image_filepaths
    else:
        raise NotADirectoryError(errno.ENOENT, os.strerror(errno.ENOENT), dir_path)


def get_dir_list_filepaths(dir_path_list, endswith_str="", startswith_str=""):
    image_filepaths = []
    for dir_path in dir_path_list:
        image_filepaths.extend(get_filepaths(dir_path, endswith_str=endswith_str, startswith_str=startswith_str))
    return image_filepaths


def save_json(filepath, data):
    dirpath = os.path.dirname(filepath)
    os.makedirs(dirpath, exist_ok=True)
    with open(filepath, 'w') as outfile:
        json.dump(data, outfile)
    return True


def load_json(filepath):
    if not os.path.exists(filepath):
        return False
    from jsmin import jsmin
    try:
        with open(filepath, 'r') as f:
            minified = jsmin(f.read())
            data = json.loads(minified)
    except json.decoder.JSONDecodeError as e:
        print_utils.print_error(f"ERROR in load_json(filepath): {e} from JSON at {filepath}")
        exit()
    return data


def wipe_dir(dirpath):
    filepaths = get_filepaths(dirpath)
    for filepath in filepaths:
        os.remove(filepath)


def split_list_into_chunks(ell, n, pad=False):
    """Yield successive n-sized chunks from ell."""
    for i in range(0, len(ell), n):
        if pad:
            chunk = ell[i:i + n]
            if len(chunk) < n:
                chunk.extend([chunk[-1]] * (n - len(chunk)))
            yield chunk
        else:
            yield ell[i:i + n]


def params_to_str(params):
    def to_str(value):
        if isinstance(value, float) and value == int(value):
            return str(int(value))
        return str(value)

    return "_".join(["{}_{}".format(key, to_str(params[key])) for key in sorted(params.keys())])


def main():
    ell = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    batches = split_list_into_chunks(ell, 4, pad=True)
    for batch in batches:
        print(batch)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
A simplified Python DVC API
"""
import ubelt as ub
import os
from kwutil import util_path
from kwutil.util_yaml import Yaml


def __test_simple_dvc():
    """
    Builds a medium complexity dvc repo, todo:
        implement some tests
    """
    import ubelt as ub
    test_root = ub.Path.appdir('simpledvc', 'tests', 'basic')
    dvc_root = test_root / 'repo'
    dvc_root.delete()
    SimpleDVC.init(dvc_root, no_scm=True)

    dvc = SimpleDVC.coerce(dvc_root)

    # Build basic data
    (dvc_root / 'test-set1').ensuredir()
    assets_dpath = (dvc_root / 'test-set1/assets').ensuredir()
    for idx in range(1, 21):
        fpath = assets_dpath / f'asset_{idx:03d}.data'
        fpath.write_text(str(idx) * 100)
    manifest_fpath = (dvc_root / 'test-set1/manifest.txt')
    manifest_fpath.write_text('pretend-data')

    root_fpath = dvc_root / 'root_file'
    root_fpath.write_text('----' * 100)

    root_dpath = dvc_root / 'root_dir'

    # Use networkx to make a random complex directory structure
    import networkx as nx
    graph = nx.erdos_renyi_graph(30, p=0.2, directed=True)
    tree = nx.minimum_spanning_arborescence(graph)
    nx.write_network_text(tree)
    sources = [n for n in tree.nodes if not tree.pred[n]]
    sinks = [n for n in tree.nodes if not tree.succ[n]]

    node_paths = []
    for t in sinks:
        for s in sources:
            paths = list(nx.all_simple_edge_paths(tree, s, t))
            if paths:
                node_path = [u for (u, v) in paths[0]] + [t]
                node_paths.append(node_path)

    for node_path in node_paths:
        rel_fpath = ub.Path(*[f'dir_{n}' for n in node_path[0:-1]]) / ('file_' + str(node_path[-1]) + '.data')
        fpath = root_dpath / rel_fpath
        fpath.parent.ensuredir()
        fpath.write_text(str(node_path))

    dvc.add(root_dpath)
    dvc.add(root_fpath)
    dvc.add(manifest_fpath)
    dvc.add(assets_dpath)

    # xdev.tree_repr(dvc_root)


class SimpleDVC(ub.NiceRepr):
    """
    A Simple DVC API

    Args:
        dvc_root (Path): path to DVC repo directory
        remote (str): dvc remote to sync to by default

    Ignore:
        >>> # xdoctest: +REQUIRES(--dvc-test)
        >>> import sys, ubelt
        >>> from geowatch.utils.simple_dvc import *  # NOQA
        >>> dvc_dpath = SimpleDVC.demo_dpath(reset=0)
        >>> self = SimpleDVC(dvc_dpath)
        >>> a_file_fpath = dvc_dpath / 'a_file.txt'
        >>> if not a_file_fpath.exists():
        >>>     a_file_fpath.write_text('hello')
        >>> self.add(a_file_fpath)

    """

    def __init__(self, dvc_root=None, remote=None):
        self.dvc_root = dvc_root
        self.remote = remote

    def __nice__(self):
        return f'dvc_root={self.dvc_root}'

    @classmethod
    def init(cls, dpath, no_scm=False, force=False, verbose=0):
        """
        Initialize a DVC repo in a path
        """
        dpath = ub.Path(dpath.ensuredir())
        args = ['dvc', 'init']
        if verbose:
            args += ['--verbose']
        if force:
            args += ['--force']
        if no_scm:
            args += ['--no-scm']
        ub.cmd(args, cwd=dpath, verbose=3, check=True)
        self = cls(dpath)
        return self

    @ub.memoize_property
    def cache_dir(self):
        info = ub.cmd('dvc cache dir', cwd=self.dvc_root, check=True)
        cache_dpath = ub.Path(info['out'].strip())
        return cache_dpath

    @classmethod
    def demo_dpath(cls, reset=False):
        dvc_dpath = ub.Path.appdir('simple_dvc/test/test_dvc_repo')
        if reset:
            dvc_dpath.delete()
        if not dvc_dpath.exists():
            dvc_dpath.ensuredir()
            verbose = 2
            # Init empty git repo
            ub.cmd('git init --quiet', cwd=dvc_dpath, verbose=verbose)
            ub.cmd('git config --local receive.denyCurrentBranch "warn"', cwd=dvc_dpath, verbose=verbose)
            # Init empty dvc repo
            ub.cmd('dvc init --quiet', cwd=dvc_dpath, verbose=verbose)
            ub.cmd('dvc config core.autostage true', cwd=dvc_dpath, verbose=verbose)
            ub.cmd('dvc config cache.type "symlink,hardlink,copy"', cwd=dvc_dpath, verbose=verbose)
            ub.cmd('dvc config cache.shared group', cwd=dvc_dpath, verbose=verbose)
            ub.cmd('dvc config cache.protected true', cwd=dvc_dpath, verbose=verbose)
        return dvc_dpath

    @classmethod
    def coerce(cls, dvc_path, **kw):
        """
        Given a path inside DVC, finds the root.
        """
        dvc_root = cls.find_root(dvc_path)
        return cls(dvc_root, **kw)

    @classmethod
    def find_root(cls, path=None):
        """
        Given a path, search its ancestors to find the root of a dvc repo.

        Returns:
            Path | None
        """
        if path is None:
            raise Exception('no way to find dvc root')
        # Need to find it from the path
        path = ub.Path(path).resolve()
        max_parts = len(path.parts)
        curr = path
        found = None
        for _ in range(max_parts + 1):
            cand = curr / '.dvc'
            if cand.exists():
                found = curr
                break
            curr = curr.parent
        return found

    def _ensure_root(self, paths):
        if self.dvc_root is None:
            self.dvc_root = self.find_root(paths[0])
            print('found new self.dvc_root = {!r}'.format(self.dvc_root))
        return self.dvc_root

    def _ensure_remote(self, remote):
        if remote is None:
            remote = self.remote
        return remote

    def _resolve_root_and_relative_paths(self, paths):
        # try:
        #     dvc_root = self._ensure_root(paths)
        #     rel_paths = [os.fspath(p.relative_to(dvc_root)) for p in paths]
        # except Exception as ex:
        #     print(f'ex={ex}')
        # Handle symlinks: https://dvc.org/doc/user-guide/troubleshooting#add-symlink
        # not sure if this is safe
        dvc_root = self._ensure_root(paths)
        if dvc_root is None:
            raise Exception('unable to find a DVC root')
        dvc_root = dvc_root.resolve()
        # Note: this could resolve the symlink to the dvc cache which we dont want
        # rel_paths = [os.fspath(p.resolve().relative_to(dvc_root)) for p in paths]
        # Fixed version?
        parent_resolved = [p.parent.resolve() / p.name for p in paths]
        rel_paths = [os.fspath(p.relative_to(dvc_root)) for p in parent_resolved]

        return dvc_root, rel_paths

    def add(self, path, verbose=0):
        """
        Args:
            path (str | PathLike | Iterable[str | PathLike]):
                a single or multiple paths to add
        """
        dvc_root, rel_paths = self._dvc_path_op('add', path, verbose)
        has_autostage = ub.cmd('dvc config core.autostage', cwd=dvc_root, check=True)['out'].strip() == 'true'
        if not has_autostage:
            print('warning: Need autostage to complete the git commit')
            # raise NotImplementedError('Need autostage to complete the git commit')

    def pathsremove(self, path, verbose=0):
        """
        Args:
            path (str | PathLike | Iterable[str | PathLike]):
                a single or multiple paths to add
        """
        self._dvc_path_op('remove', path, verbose)

    def _dvc_path_op(self, op, path, verbose=0):
        """
        Args:
            path (str | PathLike | Iterable[str | PathLike]):
                a single or multiple paths to add
        """
        dvc_main = _import_dvc_main()
        paths = list(map(ub.Path, _ensure_iterable(path)))
        if len(paths) == 0:
            print('No paths to add')
            return
        dvc_root, rel_paths = self._resolve_root_and_relative_paths(paths)
        with util_path.ChDir(dvc_root):
            dvc_command = [op] + rel_paths
            extra_args = self._verbose_extra_args(verbose)
            dvc_command = dvc_command + extra_args
            ret = dvc_main(dvc_command)
        if ret != 0:
            raise Exception(f'Failed to {op} files')
        return dvc_root, rel_paths

    def check_ignore(self, path, details=0, verbose=0):
        dvc_main = _import_dvc_main()
        paths = list(map(ub.Path, _ensure_iterable(path)))
        if len(paths) == 0:
            print('No paths to add')
            return
        dvc_root, rel_paths = self._resolve_root_and_relative_paths(paths)
        with util_path.ChDir(dvc_root):
            dvc_command = ['check-ignore'] + rel_paths
            if details:
                dvc_command += ['--details']
            extra_args = self._verbose_extra_args(verbose)
            dvc_command = dvc_command + extra_args
            ret = dvc_main(dvc_command)
        if ret != 0:
            raise Exception('Failed check-ignore')

    def git_pull(self):
        ub.cmd('git pull', verbose=3, check=True, cwd=self.dvc_root)

    def git_push(self):
        ub.cmd('git push', verbose=3, check=True, cwd=self.dvc_root)

    def git_commit(self, message):
        ub.cmd(f'git commit -m "{message}"', verbose=3, check=True, cwd=self.dvc_root)

    def git_commitpush(self, message='', pull_on_fail=True):
        """
        TODO: better name here?
        """
        # dangerous?
        try:
            self.git_commit(message)
        except Exception as e:
            ex = e
            if 'nothing added to commit' not in ex.output:
                raise
        else:
            try:
                self.git_push()
            except Exception:
                if pull_on_fail:
                    print('Initial push failed, will pull and then try once more')
                    self.git_pull()
                    self.git_push()
                else:
                    raise

    def _verbose_extra_args(self, verbose):
        extra_args = []
        if verbose:
            verbose = max(min(3, verbose), 1)
            extra_args += ['-' + 'v' * verbose]
        return extra_args

    def _remote_extra_args(self, remote, recursive, jobs, verbose):
        extra_args = self._verbose_extra_args(verbose)
        if remote:
            extra_args += ['-r', remote]
        if jobs is not None:
            extra_args += ['--jobs', str(jobs)]
        if recursive:
            extra_args += ['--recursive']
        return extra_args

    def push(self, path, remote=None, recursive=False, jobs=None, verbose=0):
        """
        Push the content tracked by .dvc files to remote storage.

        Args:
            path (Path | List[Path):
                one or more file paths that should have an associated .dvc
                sidecar file or if recursive is true, a directory containing
                multiple tracked files.

            remote (str):
                the name of the remote registered in the .dvc/config to push to

            recursive (bool):
                if True, then items in ``path`` can be a directory.

            jobs (int): number of parallel workers
        """
        dvc_main = _import_dvc_main()
        paths = list(map(ub.Path, _ensure_iterable(path)))
        if len(paths) == 0:
            print('No paths to push')
            return
        remote = self._ensure_remote(remote)
        dvc_root, rel_paths = self._resolve_root_and_relative_paths(paths)
        extra_args = self._remote_extra_args(remote, recursive, jobs, verbose)
        with util_path.ChDir(dvc_root):
            dvc_command = ['push'] + extra_args + [str(p) for p in rel_paths]
            dvc_main(dvc_command)

    def pull(self, path, remote=None, recursive=False, jobs=None, verbose=0):
        dvc_main = _import_dvc_main()
        paths = list(map(ub.Path, _ensure_iterable(path)))
        if len(paths) == 0:
            print('No paths to pull')
            return
        remote = self._ensure_remote(remote)
        dvc_root, rel_paths = self._resolve_root_and_relative_paths(paths)
        extra_args = self._remote_extra_args(remote, recursive, jobs, verbose)
        with util_path.ChDir(dvc_root):
            dvc_command = ['pull'] + extra_args + [str(p) for p in rel_paths]
            dvc_main(dvc_command)

    def request(self, path, remote=None):
        """
        Requests to ensure that a specific file from DVC exists.

        Any files that do not exist, check to see if there is an associated
        .dvc sidecar file. If any sidecar files are missing, an error is
        thrown.  Otherwise we attempt to pull the missing files.

        Args:
            path (Path | List[Path):
                one or more file paths that should have an associated .dvc
                sidecar file.
        """
        paths = list(map(ub.Path, _ensure_iterable(path)))
        missing_data = [path for path in paths if not path.exists()]

        if missing_data:
            dvc_root, rel_paths = self._resolve_root_and_relative_paths(missing_data)

            def _find_sidecar(rel_path):
                rel_path = ub.Path(rel_path)
                first_cand = dvc_root / rel_path.augment(stem=rel_path.name, ext='.dvc')
                if first_cand.exists():
                    return first_cand
                rel_parts = rel_path.parts
                for i in reversed(range(len(rel_parts))):
                    parts = rel_parts[0:i]
                    cand_dat = dvc_root.joinpath(*parts)
                    cand_dvc = cand_dat.augment(stem=cand_dat.name, ext='.dvc')
                    if cand_dvc.exists():
                        return cand_dvc
                raise IOError(f'Could not find sidecar for: {rel_path=} in {dvc_root=}. Wrong path, or need to git pull?')

            to_pull = [_find_sidecar(rel_path) for rel_path in rel_paths]
            missing_sidecar = [dvc_fpath for dvc_fpath in to_pull if not dvc_fpath.exists()]

            if missing_sidecar:
                if len(missing_sidecar) < 10:
                    print(f'missing_sidecar={missing_sidecar}')
                raise Exception(f'There were {len(missing_sidecar)} / {len(paths)} missing sidecar files')

            if to_pull:
                self.pull(to_pull, remote=remote)

    def unprotect(self, path):
        dvc_main = _import_dvc_main()
        paths = list(map(ub.Path, _ensure_iterable(path)))
        if len(paths) == 0:
            print('No paths to unprotect')
            return
        dvc_root, rel_paths = self._resolve_root_and_relative_paths(paths)
        with util_path.ChDir(dvc_root):
            dvc_command = ['unprotect'] + rel_paths
            dvc_main(dvc_command)

    def is_tracked(self, path):
        path = ub.Path(path)
        tracker_fpath = self.find_file_tracker(path)
        if tracker_fpath is not None:
            return True
        else:
            tracker_fpath = self.find_dir_tracker(path)
            if tracker_fpath is not None:
                raise NotImplementedError

    @classmethod
    def find_file_tracker(cls, path):
        assert not path.name.endswith('.dvc')
        tracker_fpath = path.augment(tail='.dvc')
        if tracker_fpath.exists():
            return tracker_fpath

    def find_dir_tracker(cls, path):
        # Find if an ancestor parent dpath is tracked
        path = ub.Path(path).absolute()
        prev = path
        dpath = path.parent
        while (not (dpath / '.dvc').exists()) and prev != dpath:
            tracker_fpath = dpath.augment(tail='.dvc')
            if tracker_fpath.exists():
                return tracker_fpath
            prev = dpath
            dpath = dpath.parent
        tracker_fpath = dpath.augment(tail='.dvc')
        if tracker_fpath.exists():
            return tracker_fpath

    def read_dvc_sidecar(self, sidecar_fpath):
        sidecar_fpath = ub.Path(sidecar_fpath)
        data = Yaml.loads(sidecar_fpath.read_text())
        return data

    def resolve_cache_paths(self, sidecar_fpath):
        """
        Given a .dvc file, enumerate the paths in the cache associated with it.

        Args:
            sidecar_fpath (PathLike | str): path to the .dvc file
        """
        sidecar_fpath = ub.Path(sidecar_fpath)
        data = Yaml.loads(sidecar_fpath.read_text())

        dvc3_cache_base = (self.cache_dir / 'files/md5')
        try_dvc3 = dvc3_cache_base.exists()

        # TODO: dvc 3.0 added new hashes! Yay! But we have to support this.
        for item in data['outs']:
            md5 = item['md5']

            if try_dvc3:
                cache_fpath = self.cache_dir / 'files' / 'md5' / md5[0:2] / md5[2:]
                if not cache_fpath.exists():
                    cache_fpath = self.cache_dir / md5[0:2] / md5[2:]
            else:
                cache_fpath = self.cache_dir / md5[0:2] / md5[2:]
                if not cache_fpath.exists():
                    cache_fpath = self.cache_dir / 'files' / 'md5' / md5[0:2] / md5[2:]

            if md5.endswith('.dir') and cache_fpath.exists():
                dir_data = Yaml.loads(cache_fpath.read_text())
                for item in dir_data:
                    file_md5 = item['md5']
                    assert not file_md5.endswith('.dir'), 'unhandled'
                    if try_dvc3:
                        file_cache_fpath = self.cache_dir / 'files' / 'md5' / file_md5[0:2] / file_md5[2:]
                    else:
                        file_cache_fpath = self.cache_dir / file_md5[0:2] / file_md5[2:]

                    yield file_cache_fpath
            yield cache_fpath

    def find_sidecar_paths(self, dpath):
        """
        Args:
            dpath (Path | str): directory in dvc repo to search

        Yields:
            ub.Path: existing dvc sidecar files
        """
        # TODO: handle .dvcignore
        dpath = ub.Path(dpath)
        for r, ds, fs in dpath.walk():
            for f in fs:
                if f.endswith('.dvc'):
                    yield r / f

    def resolve_sidecar(self, path):
        """
        Given a path in a DVC repo, resolve it to a sidecar file that it
        corresponds to. If the input is a .dvc file return it.

        If it is inside a directory that corresponds to a dvc repo, search for
        that.

        Args:
            path (Path | str): directory or file in dvc repo to search

        Yields:
            ub.Path: existing dvc sidecar files
        """
        # TODO: handle .dvcignore
        path = ub.Path(path).absolute()
        if path.name.endswith('.dvc'):
            return path
        elif path.augment(tail='.dvc').exists():
            return path.augment(tail='.dvc')
        else:
            return self.find_dir_tracker(path)


def _ensure_iterable(inputs):
    return inputs if ub.iterable(inputs) else [inputs]


####
# SimpleDVC CLI Stuff (should move to a new file)
import scriptconfig as scfg  # NOQA


class SimpleDVC_CLI(scfg.ModalCLI):
    """
    A DVC CLI That uses our simplified (and more permissive) interface.

    The main advantage is that you can run these commands outside a DVC repo as
    long as you point to a valid in-repo path.
    """

    class Add(scfg.DataConfig):
        """
        Add data to the DVC repo.
        """
        __command__ = 'add'

        paths = scfg.Value([], nargs='+', position=1, help='Input files / directories to add')

        @classmethod
        def main(cls, cmdline=1, **kwargs):
            config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
            dvc = SimpleDVC()
            dvc.add(config.paths)

    class Request(scfg.DataConfig):
        """
        Pull data if the requested file doesn't exist.
        """
        __command__ = 'request'

        paths = scfg.Value([], nargs='+', position=1, help='Data to attempt to pull')
        remote = scfg.Value(None, short_alias=['r'], help='remote to pull from if needed')

        @classmethod
        def main(cls, cmdline=1, **kwargs):
            config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
            dvc = SimpleDVC()
            dvc.request(config.paths)

    class CacheDir(scfg.DataConfig):
        """
        Print the cache directory
        """
        __command__ = 'cache_dir'

        dvc_root = scfg.Value('.', position=1, help='get the cache path for this DVC repo')

        @classmethod
        def main(cls, cmdline=1, **kwargs):
            config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
            dvc = SimpleDVC(dvc_root=config.dvc_root)
            print(dvc.cache_dir)


def _import_dvc_main():
    try:
        from dvc import main as dvc_main_mod
        dvc_main = dvc_main_mod.main
    except (ImportError, ModuleNotFoundError):
        from dvc.cli import main as dvc_main
    return dvc_main


if __name__ == '__main__':
    """

    CommandLine:
        python -m geowatch.utils.simple_dvc --help
        python -m geowatch.utils.simple_dvc request --help
        python -m geowatch.utils.simple_dvc cache_dir
    """
    SimpleDVC_CLI.main()



# flake8: noqa
from .transforms import *

from .angle_field_init import AngleFieldInit
from .approximate_polygon import ApproximatePolygon
from .filter_empty_polygons import FilterEmptyPolygons
from .filter_poly_vertex_count import FilterPolyVertexCount
from .keep_keys import KeepKeys
from .map import Map
from .tensorpoly import polygons_to_tensorpoly, tensorpoly_pad
from .tensorskeleton import Paths, Skeleton, TensorSkeleton, skeletons_to_tensorskeleton, tensorskeleton_to_skeletons
from .rasterize import Rasterize
# from .remove_doubles import RemoveDoubles
from .remove_keys import RemoveKeys
from .sample_uniform import SampleUniform
# from .to_patches import ToPatches
from .transform_by_key import TransformByKey


__all__ = [
    'functional',
    'AngleFieldInit',
    'ApproximatePolygon',
    'FilterEmptyPolygons',
    'FilterPolyVertexCount',
    'KeepKeys',
    'Map',
    'polygons_to_tensorpoly',
    'tensorpoly_pad',
    'Paths',
    'Skeleton',
    'TensorSkeleton',
    'skeletons_to_tensorskeleton',
    'tensorskeleton_to_skeletons',
    'Rasterize',
    'RemoveDoubles',
    'RemoveKeys',
    'SampleUniform',
    'ToPatches',
    'TransformByKey',
]




def test_ignore_buffer_region():
    # Basic Data Sampling
    from geowatch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
    import geowatch
    from geowatch.cli import coco_add_ignore_buffer
    import kwcoco
    import ndsampler
    import ubelt as ub
    dpath = ub.Path.appdir('geowatch/tests/ignore_buffer_2')
    dpath.ensuredir()
    ignore_buffer_size = '10@10GSD'
    dst = dpath / 'out.kwcoco.zip'
    src = geowatch.coerce_kwcoco('geowatch-msi', geodata=True, dates=True)
    kwargs = dict(src=src.data_fpath, dst=dst, ignore_buffer_size=ignore_buffer_size)

    coco_add_ignore_buffer.main(cmdline=0, **kwargs)
    coco_dset = kwcoco.CocoDataset.coerce(dst)
    sampler = ndsampler.CocoSampler(coco_dset)
    self = KWCocoVideoDataset(sampler, time_dims=4, window_dims=(300, 300),
                              channels='auto')
    self.disable_augmenter = True
    index = self.sample_grid['targets'][self.sample_grid['positives_indexes'][0]]
    item = self[index]
    # Summarize batch item in text
    summary = self.summarize_item(item)
    print('item summary: ' + ub.urepr(summary, nl=2))
    # Draw batch item
    canvas = self.draw_item(item)
    if 0:
        # xdoctest: +REQUIRES(--show)
        import kwplot
        kwplot.autompl()
        kwplot.imshow(canvas)
        kwplot.show_if_requested()

if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/tests/test_ignore_buffer_region.py
        pytest ~/code/watch/tests/test_ignore_buffer_region.py
    """
    test_ignore_buffer_region()



def format_key(key):
    if isinstance(key, list):
        for k in key:
            assert isinstance(k, str), "keys should be strings"
    else:
        assert isinstance(key, str)
    return key


class TransformByKey(object):
    """Performs data[outkey[0]], data[outkey[1]], ... = transform(data[key[0]], data[key[1]], ..)"""

    def __init__(self, transform, key=None, outkey=None, ignore_key_error=False, **kwargs):
        self.transform = transform
        if key is None:
            self.key = None
        else:
            self.key = format_key(key)
        if outkey is None:
            self.outkey = self.key
        else:
            self.outkey = format_key(outkey)
        self.ignore_key_error = ignore_key_error
        self.kwargs = kwargs

    def __call__(self, data):
        assert isinstance(data, dict), "Input data should be a dictionary, not a {}".format(type(data))
        try:
            if self.key is None:
                output = self.transform(**self.kwargs)
            elif isinstance(self.key, str):
                output = self.transform(data[self.key], **self.kwargs)
            else:
                inputs = [data[k] for k in self.key]
                output = self.transform(*inputs, **self.kwargs)

            if isinstance(self.outkey, str):
                data[self.outkey] = output
            else:
                assert isinstance(output, tuple), "Output should be tuple, not {} because outkey is {}".format(type(output), type(self.outkey))
                assert len(self.outkey) == len(output), "len(outkey) and len(output) should be the same for a 1-to-1 matching."
                for k, o in zip(self.outkey, output):
                    data[k] = o
        except KeyError as e:
            if not self.ignore_key_error:
                raise e
        return data



from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import resolve_scale_request  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import abslog_scaling  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import fliprot  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import fliprot_annot  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import inv_fliprot_annot  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import inv_fliprot  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import _string_to_hashvec  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import _boxes_snap_to_edges  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import samecolor_nodata_mask  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.data_utils import MultiscaleMask  # NOQA



import json
import os
import ubelt as ub
import scriptconfig as scfg


class FilterTracksConfig(scfg.DataConfig):
    input_kwcoco = scfg.Value(None, required=True, help=ub.paragraph(
        '''
        The input kwcoco file with the scores.
        This does not to cover all sites, any site this does not
        cover will be automatically accepted.
        '''), position=1, alias=['in_file'], group='inputs')

    threshold = scfg.Value(0.4, help=ub.paragraph(
        '''
            threshold to filter polygons, very sensitive
            '''), group='track scoring')

    input_region = scfg.Value(None, help='The coercable input region model', group='inputs')

    input_sites = scfg.Value(None, help='The coercable input site models', group='inputs')

    output_region_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        The output for the region with filtered site summaries
        '''), group='outputs')

    output_sites_dpath = scfg.Value(None, help=ub.paragraph(
        '''
        The directory where site model geojson files will be written.
        '''), alias=['out_sites_dir'], group='outputs')

    output_site_manifest_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        The file path where a manifest of all site models will be written.
        '''), alias=['out_sites_fpath'], group='outputs')


def main(cmdline=1, **kwargs):
    config = FilterTracksConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich

    rich.print('config = {}'.format(ub.urepr(config, nl=1)))

    import kwcoco
    import safer
    from geowatch.geoannots import geomodels
    from kwgis.utils import util_gis
    from kwcoco.util import util_json
    from geowatch.utils import process_context

    # Args will be serailized in kwcoco, so make sure it can be coerced to json
    jsonified_config = util_json.ensure_json_serializable(config.asdict())
    walker = ub.IndexableWalker(jsonified_config)
    for problem in util_json.find_json_unserializable(jsonified_config):
        bad_data = problem['data']
        walker[problem['loc']] = str(bad_data)

    proc_context = process_context.ProcessContext(
        name='geowatch.tasks.depth_pcd.filter_tracks', type='process',
        config=jsonified_config,
        track_emissions=False,
    )
    proc_context.start()

    # Read the region / site models we will modify
    region_model = geomodels.RegionModel.coerce(config.input_region)
    input_site_fpaths = util_gis.coerce_geojson_paths(config.input_sites)
    site_to_site_fpath = ub.udict({
        p.stem: p for p in input_site_fpaths
    })
    site_id_to_summary = ub.udict()
    for summary in region_model.site_summaries():
        assert summary.site_id not in site_id_to_summary
        site_id_to_summary[summary.site_id] = summary
    input_sites = list(geomodels.SiteModel.coerce_multiple(site_to_site_fpath.values()))

    if __debug__:
        # Check that the site paths correspond with the input site summary.
        # If they don't the following logic will produce unexpected results.
        sites_with_paths = set(site_to_site_fpath)
        sites_with_summary = set(site_id_to_summary)
        if sites_with_paths != sites_with_summary:
            print('')
            print('sites_with_paths = {}'.format(ub.urepr(sites_with_paths, nl=1)))
            print('sites_with_summary = {}'.format(ub.urepr(sites_with_summary, nl=1)))
            raise AssertionError(
                f'sites with paths {len(sites_with_paths)} are not the same as '
                f'sites with summaries {len(sites_with_summary)}')

    # Ensure caches
    for summary in site_id_to_summary.values():
        summary['properties'].setdefault('cache', {})
    for site in input_sites:
        site.header['properties'].setdefault('cache', {})

    # Build a decision and reasons for each site
    site_to_decisions = {
        s: {
            'type': 'depth_decision',
            'accept': True,
            'why': None,
            'score': None,
        }
        for s in site_id_to_summary.keys()
    }

    # track_ids_to_drop = []
    coco_dset = kwcoco.CocoDataset.coerce(config.input_kwcoco)

    # TODO: use new kwcoco track mechanisms
    tracks = coco_dset.dataset['tracks']
    tracks = [t for t in tracks if t['src'] == 'sv_depth_pcd']

    for t in tracks:
        # We are assuming track-ids correspond to site names here.
        site_id = t['id']
        if isinstance(site_id, int):
            raise NotImplementedError('new kwcoco track mechanisms')
        decision = site_to_decisions[site_id]
        accept = (t['score'] >= config.threshold)
        decision['accept'] = bool(accept)
        decision['score'] = float(t['score'])
        decision['why'] = 'threshold'

    num_accept = sum((d['accept']) for d in site_to_decisions.values())
    print(f'Filter to {num_accept} / {len(site_id_to_summary)} sites')

    # Enrich each site summary with the decision reason and update status
    for site_id, decision in site_to_decisions.items():
        sitesum = site_id_to_summary[site_id]
        # Change the status of sites to "system_rejected" instead of droping them
        if not decision['accept']:
            sitesum['properties']['status'] = 'system_rejected'
        sitesum['properties']['cache']['depth_decision'] = decision

    # Copy the site models and update their header with new summary
    # information.
    output_sites_dpath = ub.Path(config.output_sites_dpath)
    output_sites_dpath.ensuredir()
    out_site_fpaths = []

    for old_site in input_sites:
        old_fpath = site_to_site_fpath[old_site.site_id]
        new_fpath = output_sites_dpath / old_fpath.name
        new_summary = site_id_to_summary[site_id]
        old_site.header['properties']['status'] = new_summary['properties']['status']
        old_site.header['properties']['cache'].update(new_summary['properties']['cache'])
        new_fpath.write_text(old_site.dumps())
        out_site_fpaths.append(new_fpath)

    # Write the updated site summaries in a new region model
    new_summaries = list(site_id_to_summary.values())
    new_region_model = geomodels.RegionModel.from_features(
        [region_model.header] + list(new_summaries))
    output_region_fpath = ub.Path(config.output_region_fpath)
    output_region_fpath.parent.ensuredir()
    print(f'Write filtered region model to: {output_region_fpath}')
    with safer.open(output_region_fpath, 'w', temp_file=not ub.WIN32) as file:
        json.dump(new_region_model, file, indent=4)

    proc_context.stop()

    if config.output_site_manifest_fpath is not None:
        filter_output = {'type': 'tracking_result', 'info': [], 'files': [os.fspath(p) for p in out_site_fpaths]}
        filter_output['info'].append(proc_context.obj)
        print(f'Write filtered site result to {config.output_site_manifest_fpath}')
        with safer.open(config.output_site_manifest_fpath, 'w', temp_file=not ub.WIN32) as file:
            json.dump(filter_output, file, indent=4)


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements Axial-Attention layers proposed in Axial-DeepLab.

Axial-Attention factorizes 2D self-attention into two 1D self-attentions, so
that it can be applied on large inputs. Axial-Attention is typically used to
replace 3x3 convolutions in a bottleneck residual block.

[1] Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
    ECCV 2020 Spotlight.
      Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
      Liang-Chieh Chen.
"""

import numpy as np
import tensorflow as tf

from deeplab2.model import utils
from deeplab2.model.layers import activations
from deeplab2.model.layers import positional_encodings


class AxialAttention(tf.keras.layers.Layer):
    """An axial-attention layer."""

    def __init__(self,
                 query_shape=129,
                 memory_flange=32,
                 total_key_depth=512,
                 total_value_depth=1024,
                 num_heads=8,
                 name='axial_attention',
                 use_query_rpe_similarity=True,
                 use_key_rpe_similarity=True,
                 use_content_similarity=True,
                 retrieve_value_rpe=True,
                 retrieve_value_content=True,
                 initialization_std_for_query_key_rpe=1.0,
                 initialization_std_for_value_rpe=1.0,
                 self_attention_activation='softmax',
                 bn_layer=tf.keras.layers.BatchNormalization,
                 conv_kernel_weight_decay=0.0):
        """Initializes an axial-attention layer.

        This function is designed to support both global and local axial-attention
        in a unified way. If query_shape is larger than the length of input, a
        global attention is applied. If query_shape is smaller than the length of
        input, a local attention is applied. In this case, the input is divided into
        blocks of length query_shape, padded by memory_flange on both sides. Then,
        local attention is applied within each query block. The choice of
        query_shape does not affect the output value but affects computation
        efficiency and memory usage. In general, use global attention (large
        query_shape) if possible. Local axial-attention has not been supported yet.

        Args:
          query_shape: An integer, the block size for local axial attention.
            Defaults to 129 since 129 is usually the largest feature map where we do
            global attention (1025 with stride 8, or 2049 with stride 16).
          memory_flange: An integer, the memory flange padded to each query block in
            local attention. It has no effect in global attention. Defaults to 32,
            which is equivalent to a span of 65 in Aixal-DeepLab paper -- A pixel
            can see 32 pixels on its left and 32 pixels on its right.
          total_key_depth: An integer, the total depth of keys, which is also the
            depth of queries and the depth of key (query) positional encodings.
          total_value_depth: An integer, the total depth of the values, which is
            also the depth of value positional encodings.
          num_heads: An integer, the number of heads in multi-head attention.
          name: A string, the name of this axial attention layer.
          use_query_rpe_similarity: A boolean, whether to use the attention
            similarity between the queries and the relative positional encodings.
          use_key_rpe_similarity: A boolean, whether to use the attention similarity
            between the keys and the relative positional encodings.
          use_content_similarity: A boolean, whether to use the content similarity
            between the queries and the keys.
          retrieve_value_rpe: A boolean, whether to retrieve the relative positional
            encodings of the values.
          retrieve_value_content: A boolean, whether to retrieve the content of the
            values.
          initialization_std_for_query_key_rpe: A float, the initialization std for
            the relative positional encodings of the queries and keys.
          initialization_std_for_value_rpe: A float, the initialization std for the
            relative positional encodings of the values.
          self_attention_activation: A string, type of activation function for
            self-attention. Support 'sigmoid' and 'softmax'.
          bn_layer: A tf.keras.layers.Layer that computes the normalization
            (default: tf.keras.layers.BatchNormalization).
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.

        Returns:
          output: A [batch, length, total_value_depth] tensor.

        Raises:
          ValueError: If none of the three similarities (use_query_rpe_similarity,
            use_key_rpe_similarity, use_content_similarity) is used.
          ValueError: If neither of value content or value rpe is retrieved.
          ValueError: If self_attention_activation is not supported.
          ValueError: If total_key_depth is not divisible by num_heads.
          ValueError: If total_value_depth is not divisible by num_heads.
        """
        # Validate the attention similarity choices.
        if not any([
            use_content_similarity, use_key_rpe_similarity, use_query_rpe_similarity
        ]):
            raise ValueError(
                'Should use at least one similarity to compute attention.')

        # Validate the retrieve value choices.
        if not retrieve_value_content and not retrieve_value_rpe:
            raise ValueError('Should retrieve at least one of content or rpe.')

        if total_key_depth % num_heads:
            raise ValueError('Total_key_depth should be divisible by num_heads.')

        if total_value_depth % num_heads:
            raise ValueError('Total_value_depth should be divisible by num_heads.')

        super(AxialAttention, self).__init__(name=name)
        self._query_shape = query_shape
        self._memory_flange = memory_flange
        self._total_key_depth = total_key_depth
        self._total_value_depth = total_value_depth
        self._num_heads = num_heads
        self._use_query_rpe_similarity = use_query_rpe_similarity
        self._use_key_rpe_similarity = use_key_rpe_similarity
        self._use_content_similarity = use_content_similarity
        self._retrieve_value_rpe = retrieve_value_rpe
        self._retrieve_value_content = retrieve_value_content
        self._initialization_std_for_query_key_rpe = (
            initialization_std_for_query_key_rpe)
        self._initialization_std_for_value_rpe = initialization_std_for_value_rpe
        self._self_attention_activation = self_attention_activation
        self._conv_kernel_weight_decay = conv_kernel_weight_decay

        self._batch_norm_qkv = bn_layer(axis=-1, name='batch_norm_qkv')
        self._batch_norm_similarity = bn_layer(
            axis=[0, 2], name='batch_norm_similarity')
        self._batch_norm_retrieved_output = bn_layer(
            axis=[0, 2, 4], name='batch_norm_retrieved_output')

        self._key_depth_per_head = total_key_depth // num_heads
        self._attention_activate_fn = activations.get_activation(
            self_attention_activation)

    def build(self, input_shape):
        """Builds axial-attention layer weights.

        Args:
          input_shape: An integer list of length 3, the shape of the input tensor.

        Raises:
          NotImplementedError: Local axial-attention has not been implemented. It is
            triggered if query_shape is less than input_shape.
        """

        # Apply global attention if query_shape is larger than input_shape[1].
        if self._query_shape >= input_shape[1]:
            self._query_shape = input_shape[1]
            self._memory_flange = 0
        else:
            raise NotImplementedError('Local axial attention has not been '
                                      'implemented yet.')
        self._memory_shape = self._query_shape + 2 * self._memory_flange

        # Compute query key value with one convolution and an optional batch norm.
        # The initialization std is standard transformer initialization (without
        # batch norm), as used in SASA and ViT. In our case, we use batch norm by
        # default, so it does not require careful tuning. If one wants to remove
        # all batch norms in axial attention, this standard initialization should
        # still be good, but a more careful initialization is encouraged.
        self.qkv_kernel = self.add_weight(
            name='qkv_kernel',
            shape=[input_shape[-1],
                   self._total_key_depth * 2 + self._total_value_depth],
            initializer=tf.keras.initializers.TruncatedNormal(
                stddev=input_shape[-1] ** -0.5),
            regularizer=tf.keras.regularizers.l2(self._conv_kernel_weight_decay))

        if self._use_query_rpe_similarity:
            self._query_rpe = positional_encodings.RelativePositionalEncoding(
                self._query_shape,
                self._memory_shape,
                self._key_depth_per_head,
                'query_rpe',
                initialization_std=self._initialization_std_for_query_key_rpe,
                conv_kernel_weight_decay=self._conv_kernel_weight_decay)

        if self._use_key_rpe_similarity:
            self._key_rpe = positional_encodings.RelativePositionalEncoding(
                self._query_shape,
                self._memory_shape,
                self._key_depth_per_head,
                'key_rpe',
                initialization_std=self._initialization_std_for_query_key_rpe,
                conv_kernel_weight_decay=self._conv_kernel_weight_decay)

        if self._retrieve_value_rpe:
            self._value_rpe = positional_encodings.RelativePositionalEncoding(
                self._query_shape,
                self._memory_shape,
                self._total_value_depth // self._num_heads,
                'value_rpe',
                initialization_std=self._initialization_std_for_value_rpe,
                conv_kernel_weight_decay=self._conv_kernel_weight_decay)

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input [batch, length, channel] tensor.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          output: An output [batch, length, total_value_depth] tensor.
        """
        # Alternatively, the einsum can be implemented as a 1x1 convolution.
        # However, it is not obvious which implementation is more efficient (without
        # careful benchmarking), so we use einsum for its flexibility and
        # consistency with other parts of the function.
        query_key_value = tf.einsum(
            'nlc,cd->nld', input_tensor, self.qkv_kernel, name='compute_qkv')
        query_key_value = self._batch_norm_qkv(query_key_value, training=training)

        # Split query key value.
        query, key, value = tf.split(
            query_key_value,
            [self._total_key_depth, self._total_key_depth, self._total_value_depth],
            axis=-1)

        # Reshape the query, key, and value.
        query = tf.reshape(query, [-1, self._query_shape, self._num_heads,
                                   self._key_depth_per_head])
        query = tf.transpose(a=query, perm=[0, 2, 1, 3])
        key = tf.reshape(key, [-1, np.prod(self._memory_shape), self._num_heads,
                               self._key_depth_per_head])
        key = tf.transpose(a=key, perm=[0, 2, 1, 3])
        value = tf.reshape(value, [-1, np.prod(self._memory_shape), self._num_heads,
                                   self._total_value_depth // self._num_heads])

        # Gather all similarity logits into a list.
        similarity_logits = []

        # Compute the content similarity term: q * k.
        if self._use_content_similarity:
            content_similarity = tf.einsum(
                'bhld,bhmd->bhlm', query, key, name='content_similarity')
            similarity_logits.append(content_similarity)

        # Compute the query rpe similarity term: q * rpe.
        if self._use_query_rpe_similarity:
            query_rpe = self._query_rpe(None)
            query_rpe_similarity = tf.einsum(
                'bhld,lmd->bhlm', query, query_rpe, name='query_rpe_similarity')
            similarity_logits.append(query_rpe_similarity)

        # Compute the key rpe similarity term: k * rpe.
        if self._use_key_rpe_similarity:
            key_rpe = self._key_rpe(None)
            key_rpe_similarity = tf.einsum(
                'bhmd,lmd->bhlm', key, key_rpe, name='key_rpe_similarity')
            similarity_logits.append(key_rpe_similarity)

        # Apply an optional batch norm to the similarities and sum them.
        similarity_logits = tf.stack(similarity_logits)
        similarity_logits = self._batch_norm_similarity(similarity_logits,
                                                        training=training)
        similarity_logits = tf.reduce_sum(input_tensor=similarity_logits, axis=0)

        # Apply an attention activation function, e.g. softmax.
        weights = self._attention_activate_fn(similarity_logits)

        # Gather retrieved values or rpes into a list.
        retrieve_list = []

        # Retrieve the content of the attended value.
        if self._retrieve_value_content:
            retrieved_content = tf.einsum(
                'bhlm,bmhd->bhld', weights, value, name='retrieve_value_content')
            retrieve_list.append(retrieved_content)

        # Retrieve the relative position of the attended value.
        if self._retrieve_value_rpe:
            value_rpe = self._value_rpe(None)
            retrieved_rpe = tf.einsum(
                'bhlm,lmd->bhld', weights, value_rpe, name='retrieve_value_rpe')
            retrieve_list.append(retrieved_rpe)

        # Apply batch norms to retrieved contents and rpes respectively.
        retrieved_output = tf.stack(retrieve_list)
        retrieved_output = self._batch_norm_retrieved_output(retrieved_output,
                                                             training=training)
        # Additive contents and rpes.
        retrieved_output = tf.reduce_sum(input_tensor=retrieved_output, axis=0)

        # Combine the heads by transposing and reshaping the tensor.
        retrieved_output = utils.transpose_and_reshape_for_attention_operation(
            retrieved_output)

        return retrieved_output


class AxialAttention2D(tf.keras.layers.Layer):
    """Sequentially applies height-axis and width-axis axial-attention."""

    def __init__(self,
                 strides=1,
                 filters=512,
                 name='attention',
                 key_expansion=1,
                 value_expansion=2,
                 query_shape=(129, 129),
                 memory_flange=(32, 32),
                 **kwargs):
        """Initializes an AxialAttention2D layer.

        Args:
          strides: An integer, the stride for the output, usually 1 or 2.
          filters: An integer, the base number of channels for the layer.
          name: A string, the name of the attention layer.
          key_expansion: A float, the channel expansion ratio for keys.
          value_expansion: A float, the channel expansion ratio for values.
          query_shape: An integer, the maximum query shape for both the height axis
            and the width axis.
          memory_flange: An integer list of length 2. The memory flange for the
            height axis and the width axis.
          **kwargs: A dictionary of keyword arguments passed to height-axis,
            width-axis, and 2D global AxialAttention.

        Returns:
          output: A [batch, strided height, strided width, output_channels] tensor.
        """
        super(AxialAttention2D, self).__init__(name=name)
        total_key_depth = int(round(filters * key_expansion))
        total_value_depth = int(round(filters * value_expansion))
        self._strides = strides
        self._total_key_depth = total_key_depth
        self._total_value_depth = total_value_depth
        self._height_axis = AxialAttention(
            total_key_depth=total_key_depth,
            total_value_depth=total_value_depth,
            query_shape=query_shape[0],
            memory_flange=memory_flange[0],
            name='height_axis',
            **kwargs)
        self._width_axis = AxialAttention(
            total_key_depth=total_key_depth,
            total_value_depth=total_value_depth,
            query_shape=query_shape[1],
            memory_flange=memory_flange[1],
            name='width_axis',
            **kwargs)

    def call(self, inputs, training=False):
        """Performs a forward pass.

        Args:
          inputs: An input [batch, height, width, channel] tensor.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          output: An output [batch, strided_height, strided_width,
            filters * value_expansion] tensor.
        """
        _, height, width, channel = inputs.get_shape().as_list()

        # Transpose and reshape the width axis to the batch dimension.
        x = tf.transpose(a=inputs, perm=[0, 2, 1, 3])
        x = tf.reshape(x, [-1, height, channel])
        x = self._height_axis(x, training=training)
        # Reshape and transpose back to a 4D tensor.
        x = tf.reshape(x, [-1, width, height, self._total_value_depth])
        x = tf.transpose(a=x, perm=[0, 2, 1, 3])
        # Height axis striding.
        if self._strides > 1:
            x = x[:, ::self._strides, :, :]

        # Reshape the height axis to the batch dimension.
        _, strided_height, _, _ = x.get_shape().as_list()
        x = tf.reshape(x, [-1, width, self._total_value_depth])
        x = self._width_axis(x, training=training)
        # Reshape back to a 4D tensor.
        x = tf.reshape(x, [-1, strided_height, width, self._total_value_depth])
        # Width axis striding.
        if self._strides > 1:
            x = x[:, :, ::self._strides, :]

        return x


class GlobalAttention2D(tf.keras.layers.Layer):
    """A 2D global attention layer."""

    def __init__(self,
                 strides=1,
                 filters=512,
                 name='attention',
                 key_expansion=1,
                 value_expansion=2,
                 query_shape=(129, 129),
                 memory_flange=(32, 32),
                 double_global_attention=False,
                 **kwargs):
        """Initializes a GlobalAttention2D layer.

        Args:
          strides: An integer, the stride for the output, usually 1 or 2.
          filters: An integer, the base number of channels for the layer.
          name: A string, the name of the attention layer.
          key_expansion: A float, the channel expansion ratio for keys.
          value_expansion: A float, the channel expansion ratio for values.
          query_shape: An integer, the maximum query shape for both the height axis
            and the width axis.
          memory_flange: An integer list of length 2. The memory flange for the
            height axis and the width axis.
          double_global_attention: A boolean, whether to use two global attention
            layers. Two global attention layers match the parameter count to a
            seqentially applied height and width axial attention layer.
          **kwargs: A dictionary of keyword arguments passed to height-axis,
            width-axis, and 2D global AxialAttention.

        Returns:
          output: A [batch, strided height, strided width, output_channels] tensor.

        Raises:
          ValueError: If relative positional encoding is enforced in kwargs.
        """
        if any([kwargs.get('use_query_rpe_similarity', False),
                kwargs.get('use_key_rpe_similarity', False),
                kwargs.get('retrieve_value_rpe', False)]):
            raise ValueError('GlobalAttention2D does not support relative positional '
                             'encodings.')

        super(GlobalAttention2D, self).__init__(name=name)
        total_key_depth = int(round(filters * key_expansion))
        total_value_depth = int(round(filters * value_expansion))
        self._strides = strides
        self._double_global_attention = double_global_attention
        self._total_key_depth = total_key_depth
        self._total_value_depth = total_value_depth

        # Global attention does not support relative positional encodings.
        kwargs['use_query_rpe_similarity'] = False
        kwargs['use_key_rpe_similarity'] = False
        kwargs['retrieve_value_rpe'] = False
        self._kwargs = kwargs

    def build(self, input_shape):
        """Builds global attention layers according to the 4D input_shape."""
        _, height, width, _ = input_shape
        # Implement 2D global attention as 1D axial-attention by flattening the 2D
        # inputs into 1D. We also disable the relative positional encodings in
        # axial attention, so that only content-based attention is used. The query
        # shape is set to height * width, so that the axial attention is global.
        self._global = AxialAttention(
            total_key_depth=self._total_key_depth,
            total_value_depth=self._total_value_depth,
            query_shape=height * width,
            memory_flange=0,
            name='global',
            **self._kwargs)

        # Use two global attention layers in one residual block. This option
        # ensures that global attention models have similar number of layers and
        # parameters as axial-attention models.
        if self._double_global_attention:
            self._global2 = AxialAttention(
                total_key_depth=self._total_key_depth,
                total_value_depth=self._total_value_depth,
                query_shape=height * width,
                memory_flange=0,
                name='global2',
                **self._kwargs)

    def call(self, inputs, training=False):
        """Performs a forward pass.

        Args:
          inputs: An input [batch, height, width, channel] tensor.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          output: An output [batch, strided_height, strided_width,
            filters * value_expansion] tensor.
        """
        _, height, width, channel = inputs.get_shape().as_list()

        # Reshape the inputs so that the attention is global 2D.
        x = tf.reshape(inputs, [-1, height * width, channel])

        # Implement 2D global attention as 1D axial-attention by flattening the 2D
        # inputs into 1D. We also disable the relative positional encodings in
        # axial attention, so that only content-based attention is used.
        x = self._global(x, training=training)

        # Use two global attention layers in one residual block. This option
        # ensures that global attention models have the same number of layers and
        # parameters as axial-attention models.
        if self._double_global_attention:
            x = self._global2(x, training=training)
        x = tf.reshape(x, [-1, height, width, self._total_value_depth])
        if self._strides > 1:
            x = x[:, ::self._strides, ::self._strides, :]

        return x



#!/usr/bin/env python3
import scriptconfig as scfg


class PredictCLI(scfg.DataConfig):
    """
    Does not work from geowatch CLI yet. See help.
    Use ``python -m geowatch.tasks.fusion.predict`` instead for now.

    ..code:: bash

        python -m geowatch.tasks.fusion.predict --help

    Execute fusion predict
    """
    __command__ = 'predict'
    __alias__ = ['fusion_predict']

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        """
        from geowatch.tasks.fusion import predict
        return predict.main(cmdline=cmdline, **kwargs)


__cli__ = PredictCLI
main = __cli__.main

if __name__ == '__main__':
    main()



"""Code related to loading and dumping."""

import inspect
import re
from typing import Any, Callable, Dict, Tuple, Type

import yaml

from ._common import load_value_mode, parent_parser
from ._optionals import import_jsonnet, omegaconf_support
from ._type_checking import ArgumentParser

__all__ = [
    "set_loader",
    "set_dumper",
]


class DefaultLoader(getattr(yaml, "CSafeLoader", yaml.SafeLoader)):  # type: ignore[misc]
    pass


# https://stackoverflow.com/a/37958106/2732151
def remove_implicit_resolver(cls, tag_to_remove):
    if "yaml_implicit_resolvers" not in cls.__dict__:
        cls.yaml_implicit_resolvers = cls.yaml_implicit_resolvers.copy()

    for first_letter, mappings in cls.yaml_implicit_resolvers.items():
        cls.yaml_implicit_resolvers[first_letter] = [(tag, regexp) for tag, regexp in mappings if tag != tag_to_remove]


remove_implicit_resolver(DefaultLoader, "tag:yaml.org,2002:timestamp")
remove_implicit_resolver(DefaultLoader, "tag:yaml.org,2002:float")


DefaultLoader.add_implicit_resolver(
    "tag:yaml.org,2002:float",
    re.compile(
        """^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$""",
        re.X,
    ),
    list("-+0123456789."),
)


def yaml_load(stream):
    if stream.strip() == "-":
        value = stream
    else:
        value = yaml.load(stream, Loader=DefaultLoader)
    if isinstance(value, dict) and value and all(v is None for v in value.values()):
        if len(value) == 1 and stream.strip() == next(iter(value.keys())) + ":":
            value = stream
        else:
            keys = set(stream.strip(" {}").replace(" ", "").split(","))
            if len(keys) > 0 and keys == set(value.keys()):
                value = stream
    return value


def jsonnet_load(stream, path="", ext_vars=None):
    from ._jsonnet import ActionJsonnet

    ext_vars, ext_codes = ActionJsonnet.split_ext_vars(ext_vars)
    _jsonnet = import_jsonnet("jsonnet_load")
    try:
        val = _jsonnet.evaluate_snippet(path, stream, ext_vars=ext_vars, ext_codes=ext_codes)
    except RuntimeError as ex:
        try:
            return yaml_load(stream)
        except pyyaml_exceptions:
            raise ValueError(str(ex)) from ex
    return yaml_load(val)


loaders: Dict[str, Callable] = {
    "yaml": yaml_load,
    "jsonnet": jsonnet_load,
}

pyyaml_exceptions = (yaml.YAMLError,)
jsonnet_exceptions = pyyaml_exceptions + (ValueError,)

loader_exceptions: Dict[str, Tuple[Type[Exception], ...]] = {
    "yaml": pyyaml_exceptions,
    "jsonnet": jsonnet_exceptions,
}


def get_load_value_mode() -> str:
    mode = load_value_mode.get()
    if mode is None:
        parser = parent_parser.get()
        assert parser is not None
        mode = parser.parser_mode
    return mode


def get_loader_exceptions():
    return loader_exceptions[get_load_value_mode()]


def load_value(value: str, simple_types: bool = False, **kwargs):
    loader = loaders[get_load_value_mode()]
    if kwargs:
        params = set(list(inspect.signature(loader).parameters)[1:])
        kwargs = {k: v for k, v in kwargs.items() if k in params}
    loaded_value = loader(value, **kwargs)
    if not simple_types and isinstance(loaded_value, (int, float, bool, str)):
        loaded_value = value
    return loaded_value


dump_yaml_kwargs = {
    "default_flow_style": False,
    "allow_unicode": True,
    "sort_keys": False,
}

dump_json_kwargs = {
    "ensure_ascii": False,
    "sort_keys": False,
}


def yaml_dump(data):
    return yaml.safe_dump(data, **dump_yaml_kwargs)


def yaml_comments_dump(data, parser):
    dump = dumpers["yaml"](data)
    formatter = parser.formatter_class(parser.prog)
    return formatter.add_yaml_comments(dump)


def json_dump(data):
    import json

    return json.dumps(data, separators=(",", ":"), **dump_json_kwargs)


def json_indented_dump(data):
    import json

    return json.dumps(data, indent=2, **dump_json_kwargs) + "\n"


dumpers: Dict[str, Callable] = {
    "yaml": yaml_dump,
    "yaml_comments": yaml_comments_dump,
    "json": json_dump,
    "json_indented": json_indented_dump,
    "jsonnet": json_indented_dump,
}

comment_prefix: Dict[str, str] = {
    "yaml": "# ",
    "yaml_comments": "# ",
    "jsonnet": "// ",
}


def check_valid_dump_format(dump_format: str):
    if dump_format not in {"parser_mode"}.union(set(dumpers)):
        raise ValueError(f'Unknown output format "{dump_format}".')


def dump_using_format(parser: "ArgumentParser", data: dict, dump_format: str) -> str:
    if dump_format == "parser_mode":
        dump_format = parser.parser_mode if parser.parser_mode in dumpers else "yaml"
    args = (data, parser) if dump_format == "yaml_comments" else (data,)
    dump = dumpers[dump_format](*args)
    if parser.dump_header and comment_prefix.get(dump_format):
        prefix = comment_prefix[dump_format]
        header = "\n".join(prefix + line for line in parser.dump_header)
        dump = f"{header}\n{dump}"
    return dump


def set_loader(mode: str, loader_fn: Callable[[str], Any], exceptions: Tuple[Type[Exception], ...] = pyyaml_exceptions):
    """Sets the value loader function to be used when parsing with a certain mode.

    The ``loader_fn`` function must accept as input a single str type parameter
    and return any of the basic types {str, bool, int, float, list, dict, None}.
    If this function is not based on PyYAML for things to work correctly the
    exceptions types that can be raised when parsing a value fails should be
    provided.

    Args:
        mode: The parser mode for which to set its loader function. Example: "yaml".
        loader_fn: The loader function to set. Example: ``yaml.safe_load``.
        exceptions: Exceptions that the loader can raise when load fails.
            Example: (yaml.parser.ParserError, yaml.scanner.ScannerError).
    """
    loaders[mode] = loader_fn
    loader_exceptions[mode] = exceptions


def set_dumper(format_name: str, dumper_fn: Callable[[Any], str]):
    """Sets the dumping function for a given format name.

    Args:
        format_name: Name to use for dumping with this function. Example: "yaml_custom".
        dumper_fn: The dumper function to set. Example: ``yaml.safe_dump``.
    """
    dumpers[format_name] = dumper_fn


def set_omegaconf_loader():
    if omegaconf_support and "omegaconf" not in loaders:
        from ._optionals import get_omegaconf_loader

        set_loader("omegaconf", get_omegaconf_loader())



import functools
import os
import subprocess
import sys
import time

from lydorn_utils import run_utils, print_utils
from lydorn_utils import python_utils


def compute_max_disp(disp_params):
    m_g_t = disp_params["max_global_translation"]
    m_g_h = disp_params["max_global_homography"]
    m_p_t = disp_params["max_poly_translation"]
    m_p_h = disp_params["max_poly_homography"]
    m_h_c = disp_params["max_homography_coef"]
    return (m_g_t + m_h_c * m_g_h) + (m_p_t + m_h_c * m_p_h)


def get_git_revision_hash():
    try:
        hash = subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode("utf-8")[:-1]
    except subprocess.CalledProcessError:
        hash = None
    return hash


def setup_run(config):
    run_name = config["run_name"]
    new_run = config["new_run"]
    init_run_name = config["init_run_name"]

    working_dir = os.path.dirname(os.path.abspath(__file__))
    runs_dir = os.path.join(working_dir, config["runs_dirpath"])

    # setup init checkpoints directory path if one is specified:
    if init_run_name is not None:
        init_run_dirpath = run_utils.setup_run_dir(runs_dir, init_run_name)
        _, init_checkpoints_dirpath = run_utils.setup_run_subdirs(init_run_dirpath)
    else:
        init_checkpoints_dirpath = None

    # setup run directory:
    run_dirpath = run_utils.setup_run_dir(runs_dir, run_name, new_run)

    # save config in logs directory
    run_utils.save_config(config, run_dirpath)

    # save args
    args_filepath = os.path.join(run_dirpath, "args.json")
    args_to_save = {
        "run_name": run_name,
        "new_run": new_run,
        "init_run_name": init_run_name,
        "batch_size": config["optim_params"]["batch_size"],
    }
    if "samples" in config:
        args_to_save["samples"] = config["samples"]
    python_utils.save_json(args_filepath, args_to_save)

    # save current commit hash
    commit_hash = get_git_revision_hash()
    if commit_hash is not None:
        commit_hash_filepath = os.path.join(run_dirpath, "commit_history.json")
        if os.path.exists(commit_hash_filepath):
            commit_hashes = python_utils.load_json(commit_hash_filepath)
            if commit_hashes[-1] != commit_hash:
                commit_hashes.append(commit_hash)
                python_utils.save_json(commit_hash_filepath, commit_hashes)
        else:
            commit_hashes = [commit_hash]
            python_utils.save_json(commit_hash_filepath, commit_hashes)

    return run_dirpath, init_checkpoints_dirpath


def get_run_dirpath(runs_dirpath, run_name):
    working_dir = os.path.dirname(os.path.abspath(__file__))
    runs_dir = os.path.join(working_dir, runs_dirpath)

    try:
        run_dirpath = run_utils.setup_run_dir(runs_dir, run_name, check_exists=True)
    except FileNotFoundError as e:
        print_utils.print_error(f"ERROR: {e}")
        sys.exit()
    return run_dirpath


def batch_to_cuda(batch):
    # Send data to computing device:
    for key, item in batch.items():
        if hasattr(item, "cuda"):
            batch[key] = item.cuda(non_blocking=True)
    return batch


def batch_to_cpu(batch):
    # Send data to computing device:
    for key, item in batch.items():
        if hasattr(item, "cuda"):
            batch[key] = item.cpu()
    return batch


def split_batch(tile_data):
    assert len(tile_data["image"].shape) == 4, "tile_data[\"image\"] should be (N, C, H, W)"
    tile_data_list = []
    for i in range(tile_data["image"].shape[0]):
        individual_tile_data = {}
        for key, item in tile_data.items():
            if not i < len(item):
                print(key, len(item))
            individual_tile_data[key] = item[i]
        tile_data_list.append(individual_tile_data)
    return tile_data_list


def _concat_dictionaries(dict1, dict2):
    """
    Recursive concat dictionaries. Dict 1 and Dict 2 must have the same key hierarchy (this is not checked).

    :param dict1: Dictionary to add to.
    :param dict2: Dictionary to add from
    :return: Merged dictionary dict1
    """
    for key in dict1.keys():
        item1 = dict1[key]
        item2 = dict2[key]
        if isinstance(item1, dict):  # And item2 is dict too.
            dict1[key] = _concat_dictionaries(item1, item2)
        else:
            dict1[key].extend(item2)
    return dict1


def _root_concat_dictionaries(dict1, dict2):

    t0 = time.time()
    dict1 = _concat_dictionaries(dict1, dict2)
    print(f"_root_concat_dictionaries: {time.time() - t0:02}s")
    return dict1


def list_of_dicts_to_dict_of_lists(list_of_dicts):
    """
    Works recursively by using _concat_dictionaries which is recursive

    @param list_of_dicts:
    @return: dict_of_lists
    """
    return functools.reduce(_concat_dictionaries, list_of_dicts)


def flatten_dict(_dict):
    """
    Makes a hierarchy of dicts flat

    @param _dict:
    @return:
    """
    new_dict = {}
    for key, item in _dict.items():
        if isinstance(item, dict):
            item = flatten_dict(item)
            for k in item.keys():
                new_dict[key + "." + k] = item[k]
        else:
            new_dict[key] = item
    return new_dict


def _generate_list_of_dicts(list_length, methods_count, submethods_count, annotation_count, segmentation_length):
    list_of_dicts = []
    for i in range(list_length):
        d = {}
        for method_i in range(methods_count):
            d[f"method_{method_i}"] = {}
            for submethod_i in range(submethods_count):
                d[f"method_{method_i}"][f"submethod_{submethod_i}"] = []
                for annotation_i in range(annotation_count):
                    annotation = {
                        "image_id": 0,
                        "segmentation": [list(range(segmentation_length))],
                        "category_id": 100,  # Building
                        "bbox": [0, 1, 0, 1],
                        "score": 1.0
                    }
                    d[f"method_{method_i}"][f"submethod_{submethod_i}"].append(annotation)
        list_of_dicts.append(d)
    return list_of_dicts


def main():
    # list_of_dicts = [
    #     {
    #         "method1": {
    #             "submethod1": [[0, 1, 2, 3], [4, 5, 6]]
    #         }
    #     },
    #     {
    #         "method1": {
    #             "submethod1": [[7, 8, 9], [10, 11, 12, 13, 14, 15]]
    #         }
    #     },
    # ]
    t0 = time.time()
    list_of_dicts = _generate_list_of_dicts(list_length=2000, methods_count=2, submethods_count=2, annotation_count=100, segmentation_length=200)
    print(f"_generate_list_of_dicts: {time.time() - t0:02}s")

    t0 = time.time()
    dict_of_lists = list_of_dicts_to_dict_of_lists(list_of_dicts)
    print(f"list_of_dicts_to_dict_of_lists: {time.time() - t0:02}s")

    flat_dict_of_lists = flatten_dict(dict_of_lists)  # NOQA


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Iterable
from typing import TypeVar, MutableSet, Any, Generator

from torch.nn.parameter import Parameter

T = TypeVar('T')


class OrderedSet(MutableSet[T]):
    '''Ordered set using mutableset. This is necessary for having reproducible
    iteration order. This property is useful for getting reproducible
    text representation of graphviz graphs, to be used in tests. This is
    because in algorith to produce graph many set objects are iterated.'''
    def __init__(self, iterable: Any | None = None) -> None:
        self.map: dict[T, None] = {}
        if iterable is not None:
            self |= iterable

    def __len__(self) -> int:
        return len(self.map)

    def __contains__(self, value: object) -> bool:
        return value in self.map

    def add(self, value: T) -> None:
        if value not in self.map:
            self.map[value] = None

    def remove(self, value: T) -> None:
        if value in self.map:
            _ = self.map.pop(value)

    def discard(self, value: T) -> None:
        if value in self.map:
            _ = self.map.pop(value)

    def __iter__(self) -> Generator[T, Any, Any]:
        for cur in self.map:
            yield cur

    def __repr__(self) -> str:
        if not self:
            return f'{self.__class__.__name__}'
        return f'{self.__class__.__name__}({list(self)})'


def is_generator_empty(parameters: Iterable[Parameter]) -> bool:
    try:
        _ = next(iter(parameters))
        return False
    except StopIteration:
        return True


def updated_dict(
    arg_dict: dict[str, Any], update_key: str, update_value: Any
) -> dict[str, Any]:
    return {
        keyword: value if keyword != update_key else update_value
        for keyword, value in arg_dict.items()
    }


def assert_input_type(
    func_name: str, valid_input_types: tuple[type, ...], in_var: Any
) -> None:

    assert isinstance(in_var, valid_input_types), (
        f'For an unknown reason, {func_name} function was '
        f'given input with wrong type. The input is of type: '
        f'{type(in_var)}. But, it should be {valid_input_types}'
    )



"""
Is there an easy way to resample a KWCOCO regions? Like generate a 300x300
image of each project region?

Yes, here is an example where we completely rewrite everything in a fixed
resolution.
"""


def kwcoco_resample_example():
    import watch
    import numpy as np
    import kwimage
    import ubelt as ub
    import kwcoco

    USE_TOYDATA = 0
    if USE_TOYDATA:
        src_coco_dset = watch.coerce_kwcoco()
    else:
        dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        kwcoco_fpath = dvc_dpath / 'Drop4-BAS/data_vali.kwcoco.json'
        src_coco_dset = kwcoco.CocoDataset(kwcoco_fpath)

    # We will construct a new kwcoco bundle that contains the resampled data.
    # Note: resampling can be done on the fly using the methods illustrated in
    # this example, but completely rewriting the data is an extreme use case
    # that demonstrates this.
    src_coco_fpath = ub.Path(src_coco_dset.fpath)

    src_bundle_dpath = src_coco_fpath.parent
    dst_bundle_dpath = src_bundle_dpath.augment(tail='_resampled')
    dst_bundle_dpath.ensuredir()
    dst_bundle_fpath = dst_bundle_dpath / src_coco_fpath.name

    dst_coco_dset = kwcoco.CocoDataset()
    dst_coco_dset.fpath = dst_bundle_fpath

    # We want all of our videos to be this new width / height
    dst_video_dsize = np.array([300, 300])

    # Copy over categories
    for cat in src_coco_dset.index.cats.values():
        dst_coco_dset.add_category(**cat)

    # Loop through all videos / images like:
    for src_video_id in src_coco_dset.videos():

        src_video_obj: dict = src_coco_dset.index.videos[src_video_id]

        # Get the current video width / height
        src_video_dsize = np.array([src_video_obj['width'], src_video_obj['height']])

        # We need to construct a transform from the old video space to the new
        # video space. The kwimage.Affine object makes this easy.
        scale_dst_from_src = dst_video_dsize / src_video_dsize
        warp_dst_from_src = kwimage.Affine.scale(scale_dst_from_src)

        # Also get an approximate scalar scale
        approx_scale = scale_dst_from_src.mean()

        # Copy and update the video metadata
        dst_video_obj = src_video_obj.copy()
        dst_video_obj['width'] = int(dst_video_dsize[0])
        dst_video_obj['height'] = int(dst_video_dsize[1])

        # Take care when updating transform metadata. These are not kwcoco
        # standards, but we use them in watch.
        src_warp_vid_from_wld = kwimage.Affine.coerce(src_video_obj['warp_wld_to_vid'])
        dst_warp_vid_from_wld = warp_dst_from_src @ src_warp_vid_from_wld
        dst_video_obj['dst_warp_wld_to_vid'] = dst_warp_vid_from_wld.concise()
        dst_video_obj['target_gsd'] = float(src_video_obj['target_gsd'] * approx_scale)

        # Add the new video to the new coco dataset
        dst_video_id = dst_coco_dset.add_video(**dst_video_obj)

        for src_image_id in src_coco_dset.images(video_id=src_video_id):

            # Construct a CocoImage object, which will allow us to sample
            # images in any specified resolution.
            src_coco_img : kwcoco.CocoImage = src_coco_dset.coco_image(src_image_id)

            # Take care to separate categorical from continuous raster bands
            # and use proper interpolation on both.
            all_channels = src_coco_img.channels.fuse()
            categorical_bands = all_channels & kwcoco.FusedChannelSpec.coerce('quality|cloudmask')
            continuous_bands = all_channels - categorical_bands

            # Load the old delayed image in videospace and do a delayed warp
            # into the dst video / image space (which will be the same)
            src_delayed_cat = src_coco_img.imdelay(space='video', channels=categorical_bands)
            dst_delayed_cat = src_delayed_cat.warp(warp_dst_from_src, dsize=tuple(dst_video_dsize.tolist()))

            src_delayed_con = src_coco_img.imdelay(space='video', channels=continuous_bands)
            dst_delayed_con = src_delayed_con.warp(warp_dst_from_src, dsize=tuple(dst_video_dsize.tolist()))

            # Use some hacked domain knowledge to handle the right nodata value
            # (The delayed image library could be better about this).
            nodata_value = -9999
            dst_cat_im = dst_delayed_cat.finalize(nodata_method='float', antialias=False, interpolation='nearest', nodata_value=nodata_value)
            dst_con_im = dst_delayed_con.finalize(nodata_method='float', antialias=True, interpolation='cubic', nodata_value=nodata_value)

            # Reformat the data as int16 like it originally was.
            dst_con_im[np.isnan(dst_con_im)] = nodata_value
            dst_con_im = dst_con_im.astype(np.int16)
            dst_cat_im[np.isnan(dst_cat_im)] = nodata_value
            dst_cat_im = dst_cat_im.astype(np.int16)

            # Build a lookup table so we can get the raster image for any band.
            band_to_raster = {}
            for band, im in zip(categorical_bands, dst_cat_im.transpose(2, 0, 1)):
                band_to_raster[band] = im
            for band, im in zip(continuous_bands, dst_con_im.transpose(2, 0, 1)):
                band_to_raster[band] = im

            # Note: the API for constructing a new coco image with auxiliary /
            # assets objects could be improved.
            src_img_obj = src_coco_img.img
            dst_img_obj = src_img_obj.copy()
            dst_assets = dst_img_obj.pop('auxiliary')

            # Update main image metadata
            dst_valid_region = kwimage.MultiPolygon.coerce(src_img_obj['valid_region']).warp(warp_dst_from_src)
            dst_img_obj['valid_region'] = dst_valid_region.to_coco(style='new')
            dst_img_obj['width'] = int(dst_video_dsize[0])
            dst_img_obj['height'] = int(dst_video_dsize[1])

            # Can hack in an identity matrix here.
            warp_vid_from_img = kwimage.Affine.eye()
            dst_img_obj['warp_img_to_vid'] = warp_vid_from_img.concise()

            # Not sure if this propery is used anymore?
            warp_img_from_vid = warp_vid_from_img.inv()
            warp_img_from_wld = warp_img_from_vid @ dst_warp_vid_from_wld
            dst_img_obj['wld_to_pxl'] = warp_img_from_wld.concise()

            for asset in dst_assets:
                # We can hack in an identity here
                warp_img_from_asset = kwimage.Affine.eye()

                warp_asset_from_img = warp_img_from_asset.inv()
                warp_asset_from_wld = warp_asset_from_img @ warp_img_from_wld
                warp_wld_from_asset = warp_asset_from_wld.inv()

                # Update transforms in the asset / auxiliary dictionary
                asset['width'] = int(dst_video_dsize[0])
                asset['height'] = int(dst_video_dsize[1])
                asset['warp_aux_to_img'] = warp_img_from_asset.concise()
                asset['warp_to_wld'] = warp_wld_from_asset.concise()
                asset['wld_to_pxl'] = warp_asset_from_wld.concise()

                # Lookup the raster for this new image
                band = asset['channels']

                # Lookup the image data that needs to be written
                new_data = band_to_raster[band]

                # Determine where we will write it
                fname = asset['file_name']
                new_fpath = dst_bundle_dpath / fname
                old_fpath = src_bundle_dpath / fname

                assert new_fpath != old_fpath, 'dont overwrite your old bundle!'

                # We will write our new images as geotiffs and handle geotiff
                # metadata (not necessary, but makes other peoples lives
                # easier)
                write_kwargs = {}
                write_kwargs['blocksize'] = 128
                write_kwargs['compress'] = 'DEFLATE'
                write_kwargs['nodata'] = nodata_value
                from osgeo import osr
                auth = asset['wld_crs_info']['auth']
                assert auth[0] == 'EPSG', 'unhandled auth'
                epsg = auth[1]
                axis_strat = getattr(osr, asset['wld_crs_info']['axis_mapping'])
                srs = osr.SpatialReference()
                srs.ImportFromEPSG(int(epsg))
                srs.SetAxisMappingStrategy(axis_strat)
                write_kwargs['crs'] = srs.ExportToWkt()
                write_kwargs['transform'] = warp_wld_from_asset
                write_kwargs['overviews'] = 2

                # Ensure the directory structure exists and write the new asset
                new_fpath.parent.ensuredir()

                kwimage.imwrite(
                    new_fpath, new_data, backend='gdal', **write_kwargs,
                )

        # Update the image with the new video id and warped assets
        dst_img_obj['auxiliary'] = dst_assets
        dst_img_obj['video_id'] = dst_video_id

        new_image_id = dst_coco_dset.add_image(**dst_img_obj)

        # Bring over all the old annotations (although we could also just
        # reproject them from the original site model files).
        # Annots are in image space, so warp them accordingly

        src_annots = src_coco_dset.annots(gid=src_coco_img.img['id'])
        for src_annot_obj in src_annots.objs:
            dst_annot_obj = src_annot_obj.copy()

            src_poly = kwimage.MultiPolygon.coerce(dst_annot_obj['segmentation'])
            src_bbox = kwimage.Boxes([dst_annot_obj['bbox']], 'xywh')

            # We can ignore the fact that this transform is in video space
            # because we know the warp between video and image space in the new
            # dataset is the identity. A better example would account for this.
            dst_poly = src_poly.warp(warp_dst_from_src)
            dst_bbox = src_bbox.warp(warp_dst_from_src)

            dst_annot_obj['segmentation'] = dst_poly.to_coco(style='new')
            dst_annot_obj['bbox'] = list(dst_bbox.to_coco())[0]
            dst_annot_obj['image_id'] = new_image_id

            dst_coco_dset.add_annotation(**dst_annot_obj)

    # Write the new warped kwcoco file
    dst_coco_dset.dump()



r"""
Helper to fix issues in truth region / site models, particularly issues seen in
iMERIT data.

SeeAlso:
    ~/code/watch/geowatch/geoannots/geomodels.py
    ~/code/watch/geowatch/cli/validate_annotation_schemas.py
    ~/code/watch/geowatch/cli/fix_region_models.py

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    python -m geowatch.cli.fix_region_models \
        --region_models="$DVC_DATA_DPATH"/annotations/drop6/region_models/*.geojson

    python -m geowatch.cli.fix_region_models \
        --region_models "$DVC_DATA_DPATH"/submodules/annotations/region_models/*.geojson

    python -m geowatch.cli.fix_region_models \
        --region_models \
            "$DVC_DATA_DPATH"/submodules/annotations/region_models/AE_C002.geojson \
            "$DVC_DATA_DPATH"/submodules/annotations/region_models/AE_C003.geojson \
            "$DVC_DATA_DPATH"/submodules/annotations/region_models/PY_C001.geojson \
            "$DVC_DATA_DPATH"/submodules/annotations/region_models/BR_T001.geojson \
            "$DVC_DATA_DPATH"/submodules/annotations/region_models/BR_T002.geojson

    python -m geowatch.cli.validate_annotation_schemas \
        --region_models="$DVC_DATA_DPATH"/annotations/drop6/region_models/AE_C001.geojson
"""
#!/usr/bin/env python3
import decimal
import simplejson
import json
import scriptconfig as scfg
import ubelt as ub


class FixRegionModelsCLI(scfg.DataConfig):
    # site_models = scfg.Value(None, nargs='+', help='coercable site models')
    region_models = scfg.Value(None, nargs='+', help='coercable region models')


class fakefloat(float):
    def __init__(self, value):
        self._value = value

    def __repr__(self):
        return str(self._value)

    def __str__(self):
        return str(self._value)


class DecimalEncoder(json.JSONEncoder):
    # https://stackoverflow.com/questions/1960516/python-json-serialize-a-decimal-object
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return fakefloat(o)
        return super(DecimalEncoder, self).default(o)


def defaultencode(o):
    if isinstance(o, decimal.Decimal):
        # Subclass float with custom repr?
        return fakefloat(o)
    raise TypeError(repr(o) + " is not JSON serializable")


def main(cmdline=1, **kwargs):
    from kwgis.utils import util_gis
    from geowatch.geoannots import geomodels

    import rich
    config = FixRegionModelsCLI.cli(cmdline=cmdline, data=kwargs, strict=True)
    rich.print('config = ' + ub.urepr(config, nl=1))

    # dpath = '/media/joncrall/flash1/smart_data_dvc/submodules/annotations/region_models'
    region_model_fpaths = util_gis.coerce_geojson_paths(config.region_models)
    _iter = iter(region_model_fpaths)
    for fpath in _iter:
        # if fpath.stem in {'AE_C002', 'AE_C003', 'BR_T001', 'BR_T002', 'PY_C001'}:
        #     continue
        # if fpath.stem in {'AE_C002', 'AE_C003', 'PY_C001'}:
        #     continue
        region_model = geomodels.RegionModel.coerce(fpath, parse_float=decimal.Decimal)
        # region_model._validate_parts()
        try:
            region_model.validate(verbose=0)
        except Exception:
            print('Attempting a fix')
            region_model.fixup()
            fix_region_model(region_model)
            region_model.fixup()
            region_model.validate()
            if '_C' in fpath.stem:
                # Try to minimize the diff by outputing in a similar style
                new_text = special_dumps(region_model)
            else:
                # T&E regions seem to be normal json outputs
                new_text = simplejson.dumps(region_model, indent='    ')
            fpath.write_text(new_text)

        # old_text = fpath.read_text()
        # print(new_text.split('\n')[:6])
        # print(old_text.split('\n')[:6])
        # import xdev
        # print(xdev.difftext(old_text, new_text, colored=True))


def special_dumps(region_model):
    # import ubelt as ub
    t = region_model.copy()
    features = t.pop('features', None)
    lines = ['{']

    def oneline_dict(val):
        # import json
        # text = json.dumps(val, cls=DecimalEncoder)
        text = simplejson.dumps(val)
        # cls=DecimalEncoder)
        # text = ub.urepr(v, trailsep=False, nl=0)
        text = text.replace('{', '{ ')
        text = text.replace('}', ' }')
        text = text.replace('[', '[ ')
        text = text.replace(']', ' ]')
        # text = text.replace("'", '"')
        return text

    for k, v in t.items():
        text = oneline_dict(v)
        lines.append(f'"{k}": ' + text + ',')

    last_feat = features[-1]

    lines.append('"features": [')
    for feat in features:
        text = oneline_dict(feat)
        if feat is last_feat:
            lines.append(text)
        else:
            lines.append(text + ',')

    lines.append(']')
    lines.append('}')
    text = '\n'.join(lines) + '\n'
    return text


def fix_region_model(region_model):
    import mgrs
    import kwimage
    (lon,), (lat,) = region_model.geometry.centroid.xy
    mgrs_code = mgrs.MGRS().toMGRS(lat, lon, MGRSPrecision=0)

    import ubelt as ub

    DRAW_BAD_REGIONS = 0

    if DRAW_BAD_REGIONS:
        import kwplot
        kwplot.autompl()

    def draw_bad_region(region_model, region_poly):
        fig = kwplot.figure(fnum=1)
        ax = fig.gca()
        ax.cla()
        region_poly.draw(setlim=1, ax=ax, alpha=0.5)
        ax.set_title('Region Geometry: {}'.format(region_model.region_id))
        dpath = (ub.Path.home() / 'tmpfig/bad_regions').ensuredir()
        fig.savefig(dpath / region_model.region_id + '_bounds.png')

    def draw_bad_site(region_model, feat, site_poly):
        fig = kwplot.figure(fnum=1)
        ax = fig.gca()
        ax.cla()
        site_id = feat['properties']['site_id']
        site_poly.draw(setlim=1, ax=ax, alpha=0.5)
        ax.set_title('Site Geometry: {}'.format(site_id))
        dpath = (ub.Path.home() / 'tmpfig/bad_sites').ensuredir()
        fig.savefig(dpath / site_id + '_bounds.png')

    if region_model.geometry.geom_type == 'MultiPolygon':
        region_geom = region_model.geometry
        region_poly = kwimage.MultiPolygon.from_shapely(region_geom)
        parts = list(region_geom.geoms)

        if len(parts) == 0:
            raise Exception
        elif len(parts) > 1:
            # HACK! Make a convex hull!
            poly = region_poly.convex_hull
        else:
            poly = kwimage.Polygon.from_shapely(parts[0])

        if DRAW_BAD_REGIONS:
            draw_bad_region(region_model, region_poly)

        print('Fix region header geom')
        region_model.header['geometry'] = poly.to_geojson()

    if region_model.geometry.geom_type == 'Polygon':
        for ring in region_model.header['geometry']['coordinates']:
            for pt in ring:
                if len(pt) != 2:
                    assert pt[2] == 0
                    pt[:] = pt[0:2]

    very_bad_feats = []
    for feat in region_model.features:
        props = feat['properties']
        props['mgrs'] = mgrs_code

        if 'socre' in props:
            old_score = props.pop('socre', None)
            if old_score is not None:
                if 'score' not in props or props['score'] is None:
                    props['score'] = float(old_score)

        if 'score' in props:
            if isinstance(props['score'], str):
                props['score'] = float(props['score'])

        if 'model_cont' in props:
            props['model_content'] = props.pop('model_cont')

        if 'model_content' in props:
            if props['model_content'] is None:
                props['model_content'] = 'annotation'

        if 'originator' in props:
            if props['originator'] == 'imerit':
                props['originator'] = 'iMERIT'
        if 'orginator' in props:
            props['originator'] = props.pop('orginator')

        props['type'] = props['type'].replace(' ', '')

        if props['type'] == 'site_summary':
            if props['version'] is None:
                props['version'] = region_model.header['properties']['version']
            if 'version' in props:
                props['version'] = props['version'].strip()
            props['site_id'] = props['site_id'].replace(' ', '')
            props['status'] = props['status'].strip().lower()
            if 'validated' in props:
                props['validated'] = props['validated'].strip()
            if feat['geometry'] is None:
                very_bad_feats.append(feat)
            else:
                if feat['geometry']['type'] == 'MultiPolygon':
                    print('Fix site summary geom')
                    site_poly = kwimage.MultiPolygon.coerce(feat['geometry'])
                    if DRAW_BAD_REGIONS:
                        draw_bad_site(region_model, feat, site_poly)
                    parts = list(site_poly.to_shapely().geoms)
                    assert len(parts) == 1
                    poly = kwimage.Polygon.from_shapely(parts[0])
                    feat['geometry'] = poly.to_geojson()
                if 'cache' not in props:
                    props['cache'] = {}

                if feat['geometry']['type'] == 'Polygon':
                    for ring in feat['geometry']['coordinates']:
                        for pt in ring:
                            if len(pt) != 2:
                                assert pt[2] == 0
                                pt[:] = pt[0:2]

                if feat['geometry']['type'] == 'MultiPolygon':
                    for poly in feat['geometry']['coordinates']:
                        for ring in poly:
                            for pt in ring:
                                if len(pt) != 2:
                                    raise Exception
    # region_model._validate_parts()
    for feat in very_bad_feats:
        region_model['features'].remove(feat)


if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/watch/geowatch/cli/fix_region_models.py
        python -m geowatch.cli.fix_region_models
    """
    main()



"""
Utilities for rasterio

SeeAlso
    util_gdal.py
"""
import kwimage
import numpy as np
# import pygeos
import shapely as shp
import shapely.geometry
import shapely.ops
import ubelt as ub
import warnings

from contextlib import ExitStack
from dataclasses import dataclass
# from tempenv import TemporaryEnvironment
from typing import Union
# import pyproj

import rasterio
import rasterio.features
import rasterio.mask
from rasterio import Affine, MemoryFile
from rasterio.enums import Resampling

try:
    from line_profiler import profile
except Exception:
    profile = ub.identity


def _ensure_open(
        raster: Union[rasterio.DatasetReader, str]) -> rasterio.DatasetReader:
    if not isinstance(raster, rasterio.DatasetReader) or raster.closed:
        # workaround for
        # https://rasterio.readthedocs.io/en/latest/faq.html#why-can-t-rasterio-find-proj-db-rasterio-from-pypi-versions-1-2-0
        # with TemporaryEnvironment({'PROJ_LIB': None, 'PROJ_DEBUG': '3'}):
        if 1:
            return rasterio.open(raster)
    else:
        return raster


def _swapxy(poly: shp.geometry.Polygon) -> shp.geometry.Polygon:
    return kwimage.Polygon.from_shapely(poly).swap_axes().to_shapely()


@profile
def mask(raster: Union[rasterio.DatasetReader, str],
         default_nodata=None,
         save=False,
         convex_hull=False,
         as_poly=True,
         tolerance=None,
         max_polys=None,
         use_overview=0):
    """
    Compute a raster's valid data mask in pixel coordinates.

    Note that this is the rasterio mask, which for multi-band rasters is the
    binary OR of the individual band masks. This is different from the gdal
    mask, which is always per-band.

    Args:
        raster (str): Path to a dataset (raster image file)

        default_nodata (int): if raster's nodata value is None, default to this

        save (bool): if True and raster's nodata value is None, write the
            default to it. If False, performance overhead is incurred from
            creating a tempfile

        convex_hull (bool):
            if True, return the convex hull of the mask image or poly

        as_poly (bool): if True, return the mask as a shapely Polygon or
            MultiPolygon instead of a raster image, in (w, h) order (opposite
            of Python convention).

        tolerance (int): if specified, simplifies the valid polygon.

        use_overview (int):
            if non-zero uses the closest overview if it is available.
            This increases computation time, but gives a better polygon when
            use_overview is closer to 0. Note, the polygon is rescaled to
            ensure it is returned in the input pixel space, not the overview
            space.

    Returns:
        If as_poly, a shapely Polygon or MultiPolygon bounding the valid
        data region(s) in pixel coordinates.

        Else, a uint8 raster mask of the same shape as the input, where
        255 == valid and 0 == invalid.

    Ignore:
        raster = '/home/joncrall/data/dvc-repos/smart_watch_dvc/drop1/../drop1/_assets/google-cloud/LS/LC08_L1TP_016039_20160216_20170224_01_T1/LC08_L1TP_016039_20160216_20170224_01_T1_B1.TIF'  # noqa
        from geowatch.utils.util_raster import *
        mask_img = mask(raster, as_poly=False)

    Example:
        >>> # xdoctest: +REQUIRES(--network)
        >>> from geowatch.utils.util_raster import *
        >>> # FIXME; this demo path no longer has any nodata values
        >>> # Find a better demo with nodata
        >>> from geowatch.demo.landsat_demodata import grab_landsat_product
        >>> path = grab_landsat_product()['bands'][0]
        >>> #
        >>> mask_img = mask(path, as_poly=False)
        >>> import kwimage as ki
        >>> assert mask_img.shape == ki.load_image_shape(path)[:2]
        >>> got = set(np.unique(mask_img))
        >>> print(f'got={got}')
        >>> # assert got == {0, 255}  # cant do this until nodata is fixed
        >>> #
        >>> mask_poly = mask(path, as_poly=True)
        >>> import shapely
        >>> assert isinstance(mask_poly, shapely.geometry.Polygon)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.figure(fnum=1, doclf=True)
        >>> imdata = kwimage.imread(path)
        >>> imdata = kwimage.normalize_intensity(imdata)
        >>> kw_poly = kwimage.Polygon.coerce(mask_poly.buffer(0).simplify(10))
        >>> canvas = imdata.copy()
        >>> mask_alpha = kwimage.ensure_alpha_channel(mask_img, alpha=(mask_img > 0))
        >>> canvas = kwimage.overlay_alpha_layers([mask_alpha, canvas])
        >>> canvas = kw_poly.scale(0.9, about='center').draw_on(canvas, color='green', alpha=0.6)
        >>> kw_poly.scale(1.1, about='center').draw(alpha=0.5, color='red', setlim=True)
        >>> kwplot.imshow(canvas)

    Example:
        >>> # Test how the "save" functionality modifies the data
        >>> import kwimage
        >>> from geowatch.utils.util_raster import *
        >>> import pathlib
        >>> dpath = ub.Path.appdir('geowatch/tests/empty_raster').ensuredir()
        >>> raster = dpath / 'empty.tif'
        >>> ub.delete(raster)
        >>> kwimage.imwrite(raster, np.zeros((3, 3, 5)))
        >>> info1 = ub.cmd('gdalinfo {}'.format(raster))
        >>> nodata = 0
        >>> mask_img = mask(raster, as_poly=False)
        >>> print('mask_img = {!r}'.format(mask_img))
        >>> info2 = ub.cmd('gdalinfo {}'.format(raster))
        >>> mask_poly = mask(raster, as_poly=True)
        >>> info3 = ub.cmd('gdalinfo {}'.format(raster))
        >>> print(info1['out'])
        >>> print(info2['out'])
        >>> print(info3['out'])
    """
    scale_factor = None

    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', category=rasterio.errors.NotGeoreferencedWarning)
        # workaround for
        # https://rasterio.readthedocs.io/en/latest/faq.html#why-can-t-rasterio-find-proj-db-rasterio-from-pypi-versions-1-2-0

        # Do we need the temporary env anymore?
        # with TemporaryEnvironment({'PROJ_LIB': None, 'PROJ_DEBUG': '3'}):
        if True:

            img = _ensure_open(raster)
            img_height = img.height
            img_width = img.width

            # Work at the coarsest overview level for speed
            overviews = {
                tuple(img.overviews(bandx))
                for bandx in range(1, img.count + 1)
            }
            if len(overviews) == 1:
                overview_levels = ub.peek(overviews)
                if len(overview_levels):
                    img.close()
                    # Open image with a higher overview level
                    # https://github.com/rasterio/rasterio/issues/1504
                    if use_overview < 0:
                        use_overview = len(overview_levels) + use_overview

                    requested_overview = min(max(use_overview, 0), len(overview_levels) - 1)
                    img = rasterio.open(img.name,
                                        'r',
                                        overview_level=requested_overview)
                    scale_factor = overview_levels[requested_overview]

            try:
                mask_img = None
                if default_nodata is None:
                    nodata = img.nodata
                    use_disk_nodata = True
                else:
                    nodata = default_nodata
                    use_disk_nodata = False

                if nodata is None:
                    # Not specified, and not introspectable
                    # TODO: early return
                    # if as_poly:
                    #     pass
                    # else:
                    mask_img = np.full((img.height, img.width),
                                       fill_value=255,
                                       dtype=np.uint8)

                else:
                    if save:
                        raise NotImplementedError(
                            'Dont update here. It can be unsafe. '
                            'Probably should be done in a separate script')

                    # if needs_nodata:
                    #     # The image was closed, so we must open a new one
                    #     if save:
                    #         img.close()
                    #         # Add on necessary information in footer
                    #         with rasterio.open(raster, 'r+') as img:
                    #             img.nodata = nodata
                    #         img = rasterio.open(raster, 'r')
                    #     else:
                    #         profile = img.profile.copy()
                    #         profile['nodata'] = nodata
                    #         # TODO could optimize this with rasterio.shutil.copy
                    #         # or https://rasterio.readthedocs.io/en/latest/topics/windowed-rw.html#blocks
                    #         data = img.read()
                    #         img.close()
                    #         profile.update(nodata=nodata)
                    #         memfile = MemoryFile()
                    #         img = memfile.open(**profile)
                    #         img.write(data)
                    if use_disk_nodata:
                        mask_img = img.dataset_mask()
                    else:
                        # simulate 0 = nodata, 255=valid data
                        # operate inplace when possible
                        imdata = img.read(1, out_dtype=np.uint8)
                        np.not_equal(imdata,
                                     nodata,
                                     dtype=np.uint8,
                                     out=imdata)
                        np.multiply(imdata, 255, out=imdata)
                        mask_img = imdata

            finally:
                img.close()

        if convex_hull:
            from skimage.morphology import convex_hull_image
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    'ignore', 'Input image is entirely zero',
                    category=UserWarning)
                mask_img = convex_hull_image(mask_img).astype(np.uint8)

        if not as_poly:
            return mask_img

        if mask_img is None:
            raise AssertionError('mask image was None')

        # mask has values 0 and 255
        polys = []
        for poly, val in rasterio.features.shapes(mask_img, connectivity=4):
            if val > 0:
                polys.append(shp.geometry.shape(poly))
                if max_polys is not None and len(polys) > max_polys:
                    break

        if tolerance is not None:
            scaled_tolerance = tolerance
            if scale_factor is not None:
                scaled_tolerance = tolerance / scale_factor
            polys = [poly.buffer(0).simplify(scaled_tolerance) for poly in polys]

        mask_poly = shp.ops.unary_union(polys).buffer(0)

        if tolerance is not None:
            mask_poly = mask_poly.simplify(scaled_tolerance)

        # do this again to fix any weirdness from union
        if convex_hull:
            mask_poly = mask_poly.convex_hull

        if scale_factor is not None:
            # Move from area space into point space?
            # mask_poly = shapely.affinity.translate(mask_poly, xoff=-0.5, yoff=-0.5)
            import shapely.affinity
            mask_poly = shapely.affinity.scale(mask_poly,
                                               xfact=scale_factor,
                                               yfact=scale_factor,
                                               origin=(0.0, 0.0))
            # Using overviews to compute a polygon has slack.
            # Buffer to account for this.
            mask_poly = mask_poly.buffer(scale_factor)

            # Clip to the bounds
            bounds = shapely.geometry.box(0, 0, img_width, img_height)
            mask_poly = mask_poly.intersection(bounds)

            if tolerance is not None:
                mask_poly = mask_poly.simplify(tolerance)

    return mask_poly


@dataclass
class ResampledRaster(ExitStack):
    """
    Context manager to rescale a raster on the fly using rasterio

    This changes the number of pixels in the raster while maintaining its
    geographic bounds, that is, it changes the raster's GSD.

    Args:
        raster: a DatasetReader (the object returned by rasterio.open) or path
            to a dataset
        scale: factor to upscale the resolution, aka downscale the GSD, by
        read: if True, read and return the resampled data (an expensive
            operation if scale>1) else, return the resampled dataset's .profile
            attribute (metadata)
        resampling: resampling algorithm, from rasterio.enums.Resampling [1]

    Example:
        >>> # xdoctest: +REQUIRES(--slow)
        >>> # xdoctest: +REQUIRES(--network)
        >>> from geowatch.utils.util_raster import *
        >>> from geowatch.demo.landsat_demodata import grab_landsat_product
        >>> path = grab_landsat_product()['bands'][0]
        >>> #
        >>> current_gsd_meters = 60
        >>> desired_gsd_meters = 10
        >>> scale = current_gsd_meters / desired_gsd_meters
        >>> #
        >>> with rasterio.open(path) as f:
        >>>     old_profile = f.profile
        >>> #
        >>> # can instantiate this class in a with-block
        >>> with ResampledRaster(path, scale=scale, read=False) as f:
        >>>     pass
        >>> #
        >>> # or have it stick around and change the resampling on the fly
        >>> resampled = ResampledRaster(path, scale=scale, read=False)
        >>> #
        >>> # the computation only happens when you invoke 'with'
        >>> with resampled as new_profile:
        >>>     assert new_profile['width'] == int(old_profile['width'] * scale)
        >>>     assert new_profile['crs'] == old_profile['crs']
        >>> #
        >>> resampled.scale = scale / 2
        >>> resampled.read = True
        >>> #
        >>> with resampled as new:
        >>>     assert new.profile['width'] == int(old_profile['width'] * scale / 2)
        >>>     assert new.profile['crs'] == old_profile['crs']
        >>>     # do other stuff with new

    References:
        https://gis.stackexchange.com/a/329439
        https://rasterio.readthedocs.io/en/latest/topics/reading.html
        https://rasterio.readthedocs.io/en/latest/topics/profiles.html
        [1] https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling
    """
    raster: Union[str, rasterio.DatasetReader]
    scale: float = 2
    read: bool = True
    resampling: Resampling = Resampling.bilinear

    def __post_init__(self):
        super().__init__()

    def __enter__(self):
        self.raster = _ensure_open(self.raster)
        t = self.raster.transform

        # rescale the metadata
        transform = Affine(t.a / self.scale, t.b, t.c, t.d, t.e / self.scale,
                           t.f)
        height = int(np.ceil(self.raster.height * self.scale))
        width = int(np.ceil(self.raster.width * self.scale))

        profile = self.raster.profile
        profile.update(transform=transform,
                       driver='GTiff',
                       height=height,
                       width=width)

        if self.read:

            # Note changed order of indexes, arrays are band, row, col order
            # not row, col, band
            data = self.raster.read(
                out_shape=(self.raster.count, height, width),
                resampling=self.resampling,
            )

            # enter_context is from contextlib.ExitStack, which takes care of
            # closing these
            memfile = self.enter_context(MemoryFile())
            with memfile.open(**profile) as dataset:  # Open as DatasetWriter
                dataset.write(data)
                del data

            dataset = self.enter_context(
                memfile.open())  # Reopen as DatasetReader
            return dataset

        else:

            return profile

    def __exit__(self, *exc):
        pass
